/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others
Copyright (C) 2007-2008 Kristian Duske

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// models.c -- model loading and caching

// models are the only shared resource between a client and server running
// on the same machine.

#include "quakedef.h"

model_t	*loadmodel;
char	loadname[32];	// for hunk tags
char	mod_worldname[256] = {0};	// for external texture loading

void Mod_LoadSpriteModel (model_t *mod, void *buffer);
void Mod_LoadBrushModel (model_t *mod, void *buffer);
void Mod_LoadAliasModel (model_t *mod, void *buffer);
void Mod_LoadIQMModel (model_t *mod, void *buffer, char *path, qboolean crash);
qboolean Mod_FindIQMModel (model_t *mod);
model_t *Mod_LoadModel (model_t *mod, qboolean crash);

byte	mod_novis[MAX_MAP_LEAFS/8];

#define	MAX_MOD_KNOWN	2048 // johnfitz -- was 512
model_t	mod_known[MAX_MOD_KNOWN];
int		mod_numknown;

texture_t	*r_notexture_mip; // johnfitz -- moved here from r_main.c
texture_t	*r_notexture_mip2; // johnfitz -- used for non-lightmapped surfs with a missing texture

/*
===============
Mod_Init
===============
*/
void Mod_Init (void)
{
	memset (mod_novis, 0xff, sizeof (mod_novis));

	// johnfitz -- create notexture miptex
	r_notexture_mip = (texture_t *) Hunk_AllocName (sizeof (texture_t), "r_notexture_mip");
	strcpy (r_notexture_mip->name, "notexture");
	r_notexture_mip->height = r_notexture_mip->width = 32;

	r_notexture_mip2 = (texture_t *) Hunk_AllocName (sizeof (texture_t), "r_notexture_mip2");
	strcpy (r_notexture_mip2->name, "notexture2");
	r_notexture_mip2->height = r_notexture_mip2->width = 32;
	// johnfitz
}

/*
===============
Mod_Extradata

Caches the data if needed
===============
*/
void *Mod_Extradata (model_t *mod)
{
	void	*r;

	r = Cache_Check (&mod->cache);

	if (r)
		return r;

	Mod_LoadModel (mod, true);

	if (!mod->cache.data)
		Host_Error ("Mod_Extradata: caching failed");

	return mod->cache.data;
}

/*
===============
Mod_PointInLeaf
===============
*/
mleaf_t *Mod_PointInLeaf (vec3_t p, model_t *model)
{
	mnode_t		*node;
	float		d;
	mplane_t	*plane;

	if (!model || !model->nodes)
		Host_Error ("Mod_PointInLeaf: bad model");

	node = model->nodes;

	while (1)
	{
		if (node->contents < 0)
			return (mleaf_t *) node;

		plane = node->plane;
		d = DotProduct (p, plane->normal) - plane->dist;

		if (d > 0)
			node = node->children[0];
		else
			node = node->children[1];
	}

	return NULL;	// never reached
}


/*
===================
Mod_DecompressVis
===================
*/
byte *Mod_DecompressVis (byte *in, model_t *model)
{
	static byte	decompressed[MAX_MAP_LEAFS/8];
	int		c;
	byte	*out;
	int		row;

	row = (model->numleafs + 7) >> 3;
	out = decompressed;

#if 0
	memcpy (out, in, row);
#else

	if (!in)
	{
		// no vis info, so make all visible
		while (row)
		{
			*out++ = 0xff;
			row--;
		}

		return decompressed;
	}

	do
	{
		if (*in)
		{
			*out++ = *in++;
			continue;
		}

		c = in[1];
		in += 2;

		while (c)
		{
			*out++ = 0;
			c--;
		}
	}
	while (out - decompressed < row);

#endif

	return decompressed;
}

byte *Mod_LeafPVS (mleaf_t *leaf, model_t *model)
{
	if (leaf == model->leafs)
		return mod_novis;

	return Mod_DecompressVis (leaf->compressed_vis, model);
}

/*
===================
Mod_ClearAll
===================
*/
void Mod_ClearAll (void)
{
	int		i;
	model_t	*mod;

	for (i = 0, mod = mod_known; i < mod_numknown; i++, mod++)
	{
		if (mod->name[0] != '*')
		{
			// '*' models share their data with the world and so should not free it
			// everything else can safely free
			SAFE_FREE (mod->visdata);
			SAFE_FREE (mod->lightdata);
			SAFE_FREE (mod->entities);
			SAFE_FREE (mod->vertexes);
			SAFE_FREE (mod->edges);
			SAFE_FREE (mod->texinfo);
			SAFE_FREE (mod->surfaces);
			SAFE_FREE (mod->clipnodes);
			SAFE_FREE (mod->marksurfaces);
			SAFE_FREE (mod->surfedges);
			SAFE_FREE (mod->planes);
			SAFE_FREE (mod->submodels);
			SAFE_FREE (mod->hulls[0].clipnodes);
		}
		else
		{
			// in case this model slot gets reused and tries to free in the future...
			mod->visdata = NULL;
			mod->lightdata = NULL;
			mod->entities = NULL;
			mod->vertexes = NULL;
			mod->edges = NULL;
			mod->texinfo = NULL;
			mod->surfaces = NULL;
			mod->clipnodes = NULL;
			mod->marksurfaces = NULL;
			mod->surfedges = NULL;
			mod->planes = NULL;
			mod->submodels = NULL;
			mod->hulls[0].clipnodes = NULL;
		}

		if (mod->type != mod_alias)
		{
			mod->needload = true;
			TM_FreeTexturesForOwner (mod); // johnfitz
		}
	}
}

/*
==================
Mod_FindName

==================
*/
model_t *Mod_FindName (char *name)
{
	int		i;
	model_t	*mod;

	if (!name[0])
		Host_Error ("Mod_FindName: NULL name"); // johnfitz -- was "Mod_ForName"

	// search the currently loaded models
	for (i = 0, mod = mod_known; i < mod_numknown; i++, mod++)
		if (!strcmp (mod->name, name))
			break;

	if (i == mod_numknown)
	{
		if (mod_numknown == MAX_MOD_KNOWN)
			Host_Error ("mod_numknown == MAX_MOD_KNOWN");

		strcpy (mod->name, name);
		mod->needload = true;
		mod_numknown++;
	}

	return mod;
}

/*
==================
Mod_TouchModel

==================
*/
void Mod_TouchModel (char *name)
{
	model_t	*mod;

	mod = Mod_FindName (name);

	if (!mod->needload)
	{
		if (mod->type == mod_alias)
			Cache_Check (&mod->cache);
	}
}

/*
==================
Mod_LoadModel

Loads a model into the cache
==================
*/
model_t *Mod_LoadModel (model_t *mod, qboolean crash)
{
	byte	*buf;
	byte	stackbuf[1024];		// avoid dirtying the cache heap
	int	mod_type;

	if (!mod->needload)
	{
		if (mod->type == mod_alias)
		{
			if (Cache_Check (&mod->cache))
			{
				// Con_DPrintf ("retrieved %s from cache\n", mod->name);
				return mod;
			}
		}
		else return mod;		// not cached at all
	}

	// because the world is so huge, load it one piece at a time
	// load the file
	buf = COM_LoadStackFile (mod->name, stackbuf, sizeof (stackbuf));

	if (!buf)
	{
		if (crash)
			Host_Error ("Mod_LoadModel: %s not found", mod->name); // johnfitz -- was "Mod_NumForName"

		return NULL;
	}

	// allocate a new model
	COM_FileBase (mod->name, loadname);

	loadmodel = mod;

	// fill it in
	// call the apropriate loader
	mod->needload = false;

	// iqm loader...
	{
		iqmheader_t *hdr = (iqmheader_t *) buf;

		// we can only compare the magic here as the version hasn't been swapped yet
		if (!strcmp (hdr->magic, IQM_MAGIC))
		{
			Mod_LoadIQMModel (mod, buf, mod->name, crash);
			return mod;
		}
	}

	mod_type = (buf[0] | (buf[1] << 8) | (buf[2] << 16) | (buf[3] << 24));

	// this is shitty because BSP doesn't have an ident field; at least this way it's
	// a little more robust than just assuming that everything that isn't one of the
	// others is a BSP
	switch (mod_type)
	{
	case IDPOLYHEADER:
		Mod_LoadAliasModel (mod, buf);
		break;

	case IDSPRITEHEADER:
		Mod_LoadSpriteModel (mod, buf);
		break;

	case BSPVERSION:
	case BSPVERSIONRMQ:
		Mod_LoadBrushModel (mod, buf);
		break;

	default:
		Host_Error ("unknown model type %i for %s\n", mod_type, mod->name);
	}

	return mod;
}

/*
==================
Mod_ForName

Loads in a model for the given name
==================
*/
model_t *Mod_ForName (char *name, qboolean crash)
{
	model_t	*mod;

	mod = Mod_FindName (name);

	return Mod_LoadModel (mod, crash);
}


/*
===============================================================================

					BRUSHMODEL LOADING

===============================================================================
*/

byte	*mod_base;

/*
=================
Mod_CheckFullbrights -- johnfitz
=================
*/
qboolean Mod_CheckFullbrights (byte *pixels, int count)
{
	int i;

	for (i = 0; i < count; i++)
		if (*pixels++ > 223)
			return true;

	return false;
}


char bsptexturepaths[2][MAX_OSPATH];
extern gltexture_t *r_defaultnormalmap;
extern int r_normalmapregistration;

void Mod_LoadBSPTexture (texture_t *tx, miptex_t *mt, byte *srcdata)
{
	// add an additional flag so that we can run the alpha edge fix on our fence textures
	int loadflag = TEXPREF_MIPMAP | TEXPREF_OVERWRITE;
	char		texturename[64];
	src_offset_t		offset;
	int			fwidth, fheight;
	char		filename[MAX_OSPATH], filename2[MAX_OSPATH], mapname[MAX_OSPATH];
	byte		*data;

	int hunkmark = Hunk_LowMark ();

	if (tx->name[0] == '{')
	{
#if 0
		// this is testing code so that i can verify fullbrights on fences work correctly without needing to modify a map
		int r1;
		byte *pix = srcdata;

		for (r1 = 0; r1 < 128; r1++)
		{
			int r2 = rand () % (tx->width * tx->height);
			pix[r2] = 251;
		}
#endif

		loadflag |= TEXPREF_ALPHA;
	}

	if (!Q_strncasecmp (tx->name, "sky", 3))
	{
		// sky texture // also note -- was Q_strncmp, changed to match qbsp
		Sky_LoadTexture (tx, srcdata);
	}
	else
	{
		// switch the name for liquids
		if (tx->name[0] == '*') tx->name[0] = '#';

		// external textures -- first look in "textures/mapname/" then look in "textures/"
		// (added check for worldmodel using this bsp bmodel too)
		sprintf (filename, "textures/%s/%s", mod_worldname, tx->name);
		data = Image_LoadImage (filename, &fwidth, &fheight);

		if (!data)
		{
			// this could be the same as the worldmodel...
			COM_StripExtension (loadmodel->name + 5, mapname);

			// ...so only try it if it's different!
			if (strcmp (mod_worldname, mapname))
			{
				sprintf (filename, "textures/%s/%s", mapname, tx->name);
				data = Image_LoadImage (filename, &fwidth, &fheight);
			}

			if (!data)
			{
				sprintf (filename, "textures/%s", tx->name);
				data = Image_LoadImage (filename, &fwidth, &fheight);
			}
		}

		// now load whatever we found
		if (data)
		{
			// load external image
			tx->gltexture = TM_LoadImage (loadmodel, filename, fwidth, fheight, SRC_RGBA, data, filename, 0, loadflag);

			// now try to load glow/luma image from the same place
			Hunk_FreeToLowMark (hunkmark);

			// we're still in external texture land so we check #
			if (tx->name[0] != '#')
			{
				sprintf (filename2, "%s_glow", filename);
				data = Image_LoadImage (filename2, &fwidth, &fheight);

				if (!data)
				{
					sprintf (filename2, "%s_luma", filename);
					data = Image_LoadImage (filename2, &fwidth, &fheight);
				}

				if (data)
					tx->fullbright = TM_LoadImage (loadmodel, filename2, fwidth, fheight, SRC_RGBA, data, filename, 0, loadflag);
			}
		}
		else
		{
			// use the texture from the bsp file
			// need to put the old name back here first so that vid_restart will reload it properly
			if (tx->name[0] == '#') tx->name[0] = '*';

			sprintf (texturename, "%s:%s", loadmodel->name, tx->name);
			offset = (src_offset_t) (mt + 1) - (src_offset_t) mod_base;

			// we're no longer in external texture land so we check *
			if (tx->name[0] != '*' && Mod_CheckFullbrights (srcdata, (tx->width * tx->height)))
			{
				tx->gltexture = TM_LoadImage (loadmodel, texturename, tx->width, tx->height,
												  SRC_INDEXED, srcdata, loadmodel->name, offset, loadflag | TEXPREF_NOBRIGHT);
				sprintf (texturename, "%s:%s_glow", loadmodel->name, tx->name);
				tx->fullbright = TM_LoadImage (loadmodel, texturename, tx->width, tx->height,
												   SRC_INDEXED, srcdata, loadmodel->name, offset, loadflag | TEXPREF_FULLBRIGHT);
			}
			else
			{
				tx->gltexture = TM_LoadImage
				(
					loadmodel,
					texturename,
					tx->width,
					tx->height,
					SRC_INDEXED,
					srcdata,
					loadmodel->name,
					offset,
					loadflag
				);
			}
		}
	}

	// put the old name back on liquids (required for surface flags)
	if (tx->name[0] == '#') tx->name[0] = '*';

	Hunk_FreeToLowMark (hunkmark);
}


/*
=================
Mod_LoadTextures
=================
*/
void Mod_LoadTextures (lump_t *l, int version)
{
	int		i, j, num, max, altmax;
	miptex_t	*mt;
	texture_t	*tx, *tx2;
	texture_t	*anims[10];
	texture_t	*altanims[10];
	dmiptexlump_t	*m;
	// johnfitz -- more variables
	int			nummiptex;
	// johnfitz

	// johnfitz -- don't return early if no textures; still need to create dummy texture
	if (!l->filelen)
	{
		Con_Printf ("Mod_LoadTextures: no textures in bsp file\n");
		nummiptex = 0;
		m = NULL; // avoid bogus compiler warning
	}
	else
	{
		m = (dmiptexlump_t *) (mod_base + l->fileofs);
		m->nummiptex = LittleLong (m->nummiptex);
		nummiptex = m->nummiptex;
	}

	// johnfitz

	loadmodel->numtextures = nummiptex + 2; // johnfitz -- need 2 dummy texture chains for missing textures
	tx = Hunk_Alloc (loadmodel->numtextures * sizeof (texture_t));
	loadmodel->textures = (texture_t **) Hunk_Alloc (loadmodel->numtextures * sizeof (texture_t *));

	// copy out the name of the first BSP loaded in this sequence so that we can use it for subsequents
	// (e.g. external textures for BSP bmodels should be findable from the main worldmodel dir too
	if (!mod_worldname[0])
		strcpy (mod_worldname, loadname);

	// Con_Printf ("world: %s  model: %s\n", mod_worldname, loadname);

	for (i = 0; i < nummiptex; i++, tx++)
	{
		m->dataofs[i] = LittleLong (m->dataofs[i]);

		if (m->dataofs[i] == -1)
			continue;

		mt = (miptex_t *) ((byte *) m + m->dataofs[i]);
		mt->width = LittleLong (mt->width);
		mt->height = LittleLong (mt->height);

		for (j = 0; j < MIPLEVELS; j++)
			mt->offsets[j] = LittleLong (mt->offsets[j]);

		if ((mt->width & 15) || (mt->height & 15))
			Host_Error ("Texture %s is not 16 aligned", mt->name);

		// mh - storing out the pixels was horribly inefficient; they were never used in gl!
		// tx = (texture_t *) Hunk_AllocName (sizeof (texture_t), loadname); // now block-alloced above
		loadmodel->textures[i] = tx;

		memcpy (tx->name, mt->name, sizeof (tx->name));
		tx->width = mt->width;
		tx->height = mt->height;

		tx->fullbright = NULL; // johnfitz

		// johnfitz -- lots of changes
		// mh - fixme - this is gross...
		if (!isDedicated) // no texture uploading for dedicated server
		{
			Mod_LoadBSPTexture (tx, mt, (byte *) (mt + 1));
		}

		// johnfitz
	}

	// johnfitz -- last 2 slots in array should be filled with dummy textures
	loadmodel->textures[loadmodel->numtextures-2] = r_notexture_mip; // for lightmapped surfs
	loadmodel->textures[loadmodel->numtextures-1] = r_notexture_mip2; // for SURF_DRAWTILED surfs

	// sequence the animations
	for (i = 0; i < nummiptex; i++)
	{
		tx = loadmodel->textures[i];

		if (!tx || tx->name[0] != '+')
			continue;

		if (tx->anim_next)
			continue;	// allready sequenced

		// find the number of frames in the animation
		memset (anims, 0, sizeof (anims));
		memset (altanims, 0, sizeof (altanims));

		max = tx->name[1];
		altmax = 0;

		if (max >= 'a' && max <= 'z')
			max -= 'a' - 'A';

		if (max >= '0' && max <= '9')
		{
			max -= '0';
			altmax = 0;
			anims[max] = tx;
			max++;
		}
		else if (max >= 'A' && max <= 'J')
		{
			altmax = max - 'A';
			max = 0;
			altanims[altmax] = tx;
			altmax++;
		}
		else
			Host_Error ("Bad animating texture %s", tx->name);

		for (j = i + 1; j < nummiptex; j++)
		{
			tx2 = loadmodel->textures[j];

			if (!tx2 || tx2->name[0] != '+')
				continue;

			if (strcmp (tx2->name + 2, tx->name + 2))
				continue;

			num = tx2->name[1];

			if (num >= 'a' && num <= 'z')
				num -= 'a' - 'A';

			if (num >= '0' && num <= '9')
			{
				num -= '0';
				anims[num] = tx2;

				if (num + 1 > max)
					max = num + 1;
			}
			else if (num >= 'A' && num <= 'J')
			{
				num = num - 'A';
				altanims[num] = tx2;

				if (num + 1 > altmax)
					altmax = num + 1;
			}
			else
				Host_Error ("Bad animating texture %s", tx->name);
		}

#define	ANIM_CYCLE	2

		// link them all together
		for (j = 0; j < max; j++)
		{
			tx2 = anims[j];

			if (!tx2)
				Host_Error ("Missing frame %i of %s", j, tx->name);

			tx2->anim_total = max * ANIM_CYCLE;
			tx2->anim_min = j * ANIM_CYCLE;
			tx2->anim_max = (j + 1) * ANIM_CYCLE;
			tx2->anim_next = anims[(j+1) %max ];

			if (altmax)
				tx2->alternate_anims = altanims[0];
		}

		for (j = 0; j < altmax; j++)
		{
			tx2 = altanims[j];

			if (!tx2)
				Host_Error ("Missing frame %i of %s", j, tx->name);

			tx2->anim_total = altmax * ANIM_CYCLE;
			tx2->anim_min = j * ANIM_CYCLE;
			tx2->anim_max = (j + 1) * ANIM_CYCLE;
			tx2->anim_next = altanims[(j+1) %altmax ];

			if (max)
				tx2->alternate_anims = anims[0];
		}
	}
}

/*
=================
Mod_LoadLighting -- johnfitz -- replaced with lit support code via lordhavoc
=================
*/
void Mod_LoadLighting (lump_t *l, int version)
{
	int i;
	byte *in, *out, *data;
	byte d;
	char litfilename[1024];
	int hunkmark = Hunk_LowMark ();

	loadmodel->lightdata = NULL;

	// LordHavoc: check for a .lit file
	strcpy (litfilename, loadmodel->name);
	COM_StripExtension (litfilename, litfilename);
	strcat (litfilename, ".lit");
	data = (byte *) COM_LoadHunkFile (litfilename);

	// mh - lit file fix
	if (data && com_filesize == (l->filelen * 3) + 8)
	{
		if (data[0] == 'Q' && data[1] == 'L' && data[2] == 'I' && data[3] == 'T')
		{
			i = LittleLong (((int *) data)[1]);

			if (i == 1)
			{
				Con_DPrintf ("%s loaded", litfilename);
				loadmodel->lightdata = (byte *) MallocZ (l->filelen * 3);
				memcpy (loadmodel->lightdata, data + 8, l->filelen * 3);
				Hunk_FreeToLowMark (hunkmark);
				return;
			}
			else Con_Printf ("Unknown .lit file version (%d)\n", i);
		}
		else Con_Printf ("Corrupt .lit file (old version?), ignoring\n");
	}

	Hunk_FreeToLowMark (hunkmark);

	// LordHavoc: no .lit found, expand the white lighting data to color
	if (!l->filelen)
		return;

	loadmodel->lightdata = (byte *) MallocZ (l->filelen * 3);
	in = loadmodel->lightdata + l->filelen * 2; // place the file at the end, so it will not be overwritten until the very last write
	out = loadmodel->lightdata;
	memcpy (in, mod_base + l->fileofs, l->filelen);

	for (i = 0; i < l->filelen; i++)
	{
		d = *in++;
		*out++ = d;
		*out++ = d;
		*out++ = d;
	}
}


/*
=================
Mod_LoadVisibility
=================
*/
void Mod_LoadVisibility (lump_t *l, int version)
{
	if (!l->filelen)
	{
		loadmodel->visdata = NULL;
		return;
	}

	loadmodel->visdata = (byte *) MallocZ (l->filelen);
	memcpy (loadmodel->visdata, mod_base + l->fileofs, l->filelen);
}


/*
=================
Mod_LoadEntities
=================
*/
void Mod_LoadEntities (lump_t *l, int version)
{
	if (!l->filelen)
	{
		loadmodel->entities = NULL;
		return;
	}

	loadmodel->entities = (char *) MallocZ (l->filelen);
	memcpy (loadmodel->entities, mod_base + l->fileofs, l->filelen);
}


/*
=================
Mod_LoadVertexes
=================
*/
void Mod_LoadVertexes (lump_t *l, int version)
{
	dvertex_t	*in;
	mvertex_t	*out;
	int			i, count;

	in = (dvertex_t *) (mod_base + l->fileofs);

	if (l->filelen % sizeof (dvertex_t))
		Host_Error ("Mod_LoadVertexes: funny lump size in %s", loadmodel->name);

	count = l->filelen / sizeof (dvertex_t);

	out = (mvertex_t *) MallocZ (count * sizeof (mvertex_t));

	loadmodel->vertexes = out;
	loadmodel->numvertexes = count;

	for (i = 0; i < count; i++, in++, out++)
	{
		out->position[0] = LittleFloat (in->point[0]);
		out->position[1] = LittleFloat (in->point[1]);
		out->position[2] = LittleFloat (in->point[2]);

		out->dlightframe = -1;
		out->dlrgb[0] = 0;
		out->dlrgb[1] = 0;
		out->dlrgb[2] = 0;
	}
}

/*
=================
Mod_LoadEdges
=================
*/
void Mod_LoadEdges (lump_t *l, int version)
{
	if (version == BSPVERSIONRMQ)
	{
		dedge29a_t *in;
		medge_t *out;
		int	i, count;

		in = (dedge29a_t *) (mod_base + l->fileofs);

		if (l->filelen % sizeof (*in))
			Host_Error ("Mod_LoadEdges: funny lump size in %s", loadmodel->name);

		count = l->filelen / sizeof (*in);
		out = (medge_t *) MallocZ ((count + 1) * sizeof (*out));

		loadmodel->edges = out;
		loadmodel->numedges = count;

		for (i = 0; i < count; i++, in++, out++)
		{
			out->v[0] = LittleLong (in->v[0]);
			out->v[1] = LittleLong (in->v[1]);
		}
	}
	else
	{
		dedge_t *in;
		medge_t *out;
		int	i, count;

		in = (dedge_t *) (mod_base + l->fileofs);

		if (l->filelen % sizeof (*in))
			Host_Error ("Mod_LoadEdges: funny lump size in %s", loadmodel->name);

		count = l->filelen / sizeof (*in);
		out = (medge_t *) MallocZ ((count + 1) * sizeof (*out));

		loadmodel->edges = out;
		loadmodel->numedges = count;

		for (i = 0; i < count; i++, in++, out++)
		{
			out->v[0] = (unsigned short) LittleShort (in->v[0]);
			out->v[1] = (unsigned short) LittleShort (in->v[1]);
		}
	}
}


/*
=================
Mod_LoadTexinfo
=================
*/
void Mod_LoadTexinfo (lump_t *l, int version)
{
	texinfo_t *in;
	mtexinfo_t *out;
	int	i, j, count, miptex;
	float	len1, len2;
	int missing = 0; // johnfitz

	in = (texinfo_t *) (mod_base + l->fileofs);

	if (l->filelen % sizeof (*in))
		Host_Error ("Mod_LoadTexinfo: funny lump size in %s", loadmodel->name);

	count = l->filelen / sizeof (*in);
	out = (mtexinfo_t *) MallocZ (count * sizeof (*out));

	loadmodel->texinfo = out;
	loadmodel->numtexinfo = count;

	for (i = 0; i < count; i++, in++, out++)
	{
		for (j = 0; j < 8; j++)
			out->vecs[0][j] = LittleFloat (in->vecs[0][j]);

		len1 = Length (out->vecs[0]);
		len2 = Length (out->vecs[1]);
		len1 = (len1 + len2) / 2;

		if (len1 < 0.32)
			out->mipadjust = 4;
		else if (len1 < 0.49)
			out->mipadjust = 3;
		else if (len1 < 0.99)
			out->mipadjust = 2;
		else
			out->mipadjust = 1;

#if 0

		if (len1 + len2 < 0.001)
			out->mipadjust = 1;		// don't crash
		else
			out->mipadjust = 1 / floor ((len1 + len2) / 2 + 0.1);

#endif

		miptex = LittleLong (in->miptex);
		out->flags = LittleLong (in->flags);

		// johnfitz -- rewrote this section
		if (miptex >= loadmodel->numtextures - 1 || !loadmodel->textures[miptex])
		{
			if (out->flags & TEX_SPECIAL)
				out->texture = loadmodel->textures[loadmodel->numtextures-1];
			else
				out->texture = loadmodel->textures[loadmodel->numtextures-2];

			out->flags |= TEX_MISSING;
			missing++;
		}
		else
		{
			out->texture = loadmodel->textures[miptex];
		}

		// johnfitz
	}

	// johnfitz: report missing textures
	if (missing && loadmodel->numtextures > 1)
		Con_Printf ("Mod_LoadTexinfo: %d texture(s) missing from BSP file\n", missing);

	// johnfitz
}

/*
================
CalcSurfaceExtents

Fills in s->texturemins[] and s->extents[]
================
*/
void CalcSurfaceExtents (msurface_t *s)
{
	float	mins[2], maxs[2], val;
	int		i, j, e;
	mvertex_t	*v;
	mtexinfo_t	*tex;
	int		bmins[2], bmaxs[2];

	mins[0] = mins[1] = 999999;
	maxs[0] = maxs[1] = -99999;

	tex = s->texinfo;

	for (i = 0; i < s->numedges; i++)
	{
		e = loadmodel->surfedges[s->firstedge+i];

		if (e >= 0)
			v = &loadmodel->vertexes[loadmodel->edges[e].v[0]];
		else v = &loadmodel->vertexes[loadmodel->edges[-e].v[1]];

		for (j = 0; j < 2; j++)
		{
			val = v->position[0] * tex->vecs[j][0] +
				  v->position[1] * tex->vecs[j][1] +
				  v->position[2] * tex->vecs[j][2] +
				  tex->vecs[j][3];

			if (val < mins[j])
				mins[j] = val;

			if (val > maxs[j])
				maxs[j] = val;
		}
	}

	for (i = 0; i < 2; i++)
	{
		bmins[i] = floor (mins[i] / 16);
		bmaxs[i] = ceil (maxs[i] / 16);

		s->texturemins[i] = bmins[i] * 16;
		s->extents[i] = (bmaxs[i] - bmins[i]) * 16;

		// mh - new max extent; (lm block size - 1) << 4
		if (!(tex->flags & TEX_SPECIAL) && s->extents[i] > 8176) // johnfitz -- was 512 in glquake, 256 in winquake
			Host_Error ("CalcSurfaceExtents : Bad surface extents");
	}
}


/*
=================
Mod_CalcSurfaceBounds -- johnfitz -- calculate bounding box for per-surface frustum culling
=================
*/
void Mod_CalcSurfaceBounds (msurface_t *s)
{
	int			i, e;
	mvertex_t	*v;

	s->mins[0] = s->mins[1] = s->mins[2] = 9999999;
	s->maxs[0] = s->maxs[1] = s->maxs[2] = -9999999;

	// cache base verts
	s->baseverts = (mvertex_t **) Hunk_Alloc (s->numedges * sizeof (mvertex_t *));

	for (i = 0; i < s->numedges; i++)
	{
		e = loadmodel->surfedges[s->firstedge+i];

		if (e >= 0)
			v = &loadmodel->vertexes[loadmodel->edges[e].v[0]];
		else v = &loadmodel->vertexes[loadmodel->edges[-e].v[1]];

		if (s->mins[0] > v->position[0]) s->mins[0] = v->position[0];
		if (s->mins[1] > v->position[1]) s->mins[1] = v->position[1];
		if (s->mins[2] > v->position[2]) s->mins[2] = v->position[2];
		if (s->maxs[0] < v->position[0]) s->maxs[0] = v->position[0];
		if (s->maxs[1] < v->position[1]) s->maxs[1] = v->position[1];
		if (s->maxs[2] < v->position[2]) s->maxs[2] = v->position[2];

		// cache base verts
		s->baseverts[i] = v;
	}
}


/*
=================
Mod_LoadFaces
=================
*/
void Mod_LoadFacesInitCommon (lump_t *l, msurface_t **out, int *count, int insize, int outsize)
{
	if (l->filelen % insize)
		Host_Error ("Mod_LoadFacesInitCommon: funny lump size in %s", loadmodel->name);

	*count = l->filelen / insize;
	*out = (msurface_t *) MallocZ (*count * outsize);

	// johnfitz -- warn mappers about exceeding old limits
	if (*count > 32767)
		Con_DevWarning ("%i faces exceeds standard limit of 32767.\n", *count);

	loadmodel->surfaces = *out;
	loadmodel->numsurfaces = *count;
}


void Mod_LoadFacesPerFace1Common (msurface_t *out)
{
	// setup glvertexes and glindexes
	out->numvertexes = out->numedges;
	out->numindexes = (out->numvertexes - 2) * 3;

	// store out the model that uses this surf so that we can easily regenerate verts if required
	out->model = loadmodel;
}


void Mod_SetFaceFlags (msurface_t *out)
{
	if (out->texinfo->texture->name[0] == '*') // warp surface
	{
		out->flags |= (SURF_DRAWTURB | SURF_DRAWTILED);

		// detect special liquid types
		if (!strncmp (out->texinfo->texture->name, "*lava", 5))
			out->flags |= SURF_DRAWLAVA;
		else if (!strncmp (out->texinfo->texture->name, "*slime", 6))
			out->flags |= SURF_DRAWSLIME;
		else if (!strncmp (out->texinfo->texture->name, "*tele", 5))
			out->flags |= SURF_DRAWTELE;
		else out->flags |= SURF_DRAWWATER;
	}
	else if (out->texinfo->texture->name[0] == '{')
	{
		out->flags |= SURF_DRAWFENCE;
	}
	else if (!Q_strncasecmp (out->texinfo->texture->name, "sky", 3))
	{
		// oops!!!
		out->flags |= (SURF_DRAWSKY | SURF_DRAWTILED);
	}
}


void Mod_LoadFaces (lump_t *l, int version)
{
	if (version == BSPVERSIONRMQ)
	{
		dface29a_t	*in;
		msurface_t 	*out;
		int			i, count, surfnum;
		int			planenum, side;

		in = (dface29a_t *) (mod_base + l->fileofs);
		Mod_LoadFacesInitCommon (l, &out, &count, sizeof (dface29a_t), sizeof (msurface_t));

		for (surfnum = 0; surfnum < count; surfnum++, in++, out++)
		{
			out->surfnum = surfnum;
			out->firstedge = LittleLong (in->firstedge);
			out->numedges = LittleLong (in->numedges);
			out->flags = 0;

			out->texinfo = loadmodel->texinfo + LittleLong (in->texinfo);
			Mod_LoadFacesPerFace1Common (out);

			planenum = LittleLong (in->planenum);
			side = LittleLong (in->side);
			if (side) out->flags |= SURF_PLANEBACK;
			out->plane = loadmodel->planes + planenum;

			CalcSurfaceExtents (out);
			Mod_CalcSurfaceBounds (out); // johnfitz -- for per-surface frustum culling

			// lighting info
			for (i = 0; i < MAX_SURFACE_STYLES; i++)
				out->styles[i] = in->styles[i];

			i = LittleLong (in->lightofs);

			if (i == -1)
				out->samples = NULL;
			else out->samples = loadmodel->lightdata + (i * 3); // johnfitz -- lit support via lordhavoc (was "+ i")

			Mod_SetFaceFlags (out);
		}
	}
	else
	{
		dface_t		*in;
		msurface_t 	*out;
		int			i, count, surfnum;
		int			planenum, side;

		in = (dface_t *) (mod_base + l->fileofs);
		Mod_LoadFacesInitCommon (l, &out, &count, sizeof (dface_t), sizeof (msurface_t));

		for (surfnum = 0; surfnum < count; surfnum++, in++, out++)
		{
			out->surfnum = surfnum;
			out->firstedge = LittleLong (in->firstedge);
			out->numedges = LittleShort (in->numedges);
			out->flags = 0;

			out->texinfo = loadmodel->texinfo + LittleShort (in->texinfo);
			Mod_LoadFacesPerFace1Common (out);

			planenum = LittleShort (in->planenum);
			side = LittleShort (in->side);
			if (side) out->flags |= SURF_PLANEBACK;
			out->plane = loadmodel->planes + planenum;

			CalcSurfaceExtents (out);
			Mod_CalcSurfaceBounds (out); // johnfitz -- for per-surface frustum culling

			// lighting info
			for (i = 0; i < MAX_SURFACE_STYLES; i++)
				out->styles[i] = in->styles[i];

			i = LittleLong (in->lightofs);

			if (i == -1)
				out->samples = NULL;
			else out->samples = loadmodel->lightdata + (i * 3); // johnfitz -- lit support via lordhavoc (was "+ i")

			Mod_SetFaceFlags (out);
		}
	}
}


/*
=================
Mod_SetParent
=================
*/
void Mod_SetParent (mnode_t *node, mnode_t *parent)
{
	node->parent = parent;

	if (node->contents < 0)
		return;

	Mod_SetParent (node->children[0], node);
	Mod_SetParent (node->children[1], node);
}

/*
=================
Mod_LoadNodes
=================
*/
void Mod_LoadNodes (lump_t *l, int version)
{
	if (version == BSPVERSIONRMQ)
	{
		int			i, j, count, p;
		dnode29a_t	*in;
		mnode_t		*out;

		in = (dnode29a_t *) (mod_base + l->fileofs);

		if (l->filelen % sizeof (*in))
			Host_Error ("Mod_LoadNodes: funny lump size in %s", loadmodel->name);

		count = l->filelen / sizeof (*in);
		out = (mnode_t *) Hunk_Alloc (count * sizeof (*out));

		loadmodel->nodes = out;
		loadmodel->numnodes = count;

		for (i = 0; i < count; i++, in++, out++)
		{
			for (j = 0; j < 3; j++)
			{
				out->minmaxs[j] = LittleShort (in->mins[j]);
				out->minmaxs[3+j] = LittleShort (in->maxs[j]);
			}

			p = LittleLong (in->planenum);
			out->plane = loadmodel->planes + p;

			out->firstsurface = LittleLong (in->firstface);
			out->numsurfaces = LittleLong (in->numfaces);

			for (j = 0; j < 2; j++)
			{
				p = LittleLong (in->children[j]);

				if (p >= 0)
					out->children[j] = loadmodel->nodes + p;
				else out->children[j] = (mnode_t *) (loadmodel->leafs + (-1 - p));
			}
		}
	}
	else
	{
		int			i, j, count, p;
		dnode_t		*in;
		mnode_t		*out;

		in = (dnode_t *) (mod_base + l->fileofs);

		if (l->filelen % sizeof (*in))
			Host_Error ("Mod_LoadNodes: funny lump size in %s", loadmodel->name);

		count = l->filelen / sizeof (*in);
		out = (mnode_t *) Hunk_AllocName (count * sizeof (*out), loadname);

		// johnfitz -- warn mappers about exceeding old limits
		if (count > 32767)
			Con_DevWarning ("%i nodes exceeds standard limit of 32767.\n", count);
		// johnfitz

		loadmodel->nodes = out;
		loadmodel->numnodes = count;

		for (i = 0; i < count; i++, in++, out++)
		{
			for (j = 0; j < 3; j++)
			{
				out->minmaxs[j] = LittleShort (in->mins[j]);
				out->minmaxs[3+j] = LittleShort (in->maxs[j]);
			}

			p = LittleLong (in->planenum);
			out->plane = loadmodel->planes + p;

			out->firstsurface = (unsigned short) LittleShort (in->firstface); // johnfitz -- explicit cast as unsigned short
			out->numsurfaces = (unsigned short) LittleShort (in->numfaces); // johnfitz -- explicit cast as unsigned short

			for (j = 0; j < 2; j++)
			{
				// johnfitz -- hack to handle nodes > 32k, adapted from darkplaces
				p = (unsigned short) LittleShort (in->children[j]);

				if (p < count)
					out->children[j] = loadmodel->nodes + p;
				else
				{
					p = 65535 - p; // note this uses 65535 intentionally, -1 is leaf 0

					if (p < loadmodel->numleafs)
						out->children[j] = (mnode_t *) (loadmodel->leafs + p);
					else
					{
						Con_Printf ("Mod_LoadNodes: invalid leaf index %i (file has only %i leafs)\n", p, loadmodel->numleafs);
						out->children[j] = (mnode_t *) (loadmodel->leafs); // map it to the solid leaf
					}
				}
				// johnfitz
			}
		}
	}

	Mod_SetParent (loadmodel->nodes, NULL);	// sets nodes and leafs
}


/*
=================
Mod_LoadLeafs
=================
*/
void Mod_LoadLeafs (lump_t *l, int version)
{
	if (version == BSPVERSIONRMQ)
	{
		dleaf29a_t	*in;
		mleaf_t		*out;
		int			i, j, count, p;

		in = (dleaf29a_t *) (mod_base + l->fileofs);

		if (l->filelen % sizeof (*in))
			Host_Error ("Mod_LoadLeafs: funny lump size in %s", loadmodel->name);

		count = l->filelen / sizeof (*in);
		out = (mleaf_t *) Hunk_AllocName (count * sizeof (*out), loadname);

		loadmodel->leafs = out;
		loadmodel->numleafs = count;

		for (i = 0; i < count; i++, in++, out++)
		{
			for (j = 0; j < 3; j++)
			{
				out->minmaxs[j] = LittleShort (in->mins[j]);
				out->minmaxs[3+j] = LittleShort (in->maxs[j]);
			}

			p = LittleLong (in->contents);
			out->contents = p;

			out->firstmarksurface = loadmodel->marksurfaces + LittleLong (in->firstmarksurface);
			out->nummarksurfaces = LittleLong (in->nummarksurfaces);

			p = LittleLong (in->visofs);

			if (p == -1)
				out->compressed_vis = NULL;
			else
				out->compressed_vis = loadmodel->visdata + p;

			out->efrags = NULL;

			for (j = 0; j < 4; j++)
				out->ambient_sound_level[j] = in->ambient_level[j];

			// johnfitz -- removed code to mark surfaces as SURF_UNDERWATER
		}
	}
	else
	{
		dleaf_t		*in;
		mleaf_t		*out;
		int			i, j, count, p;

		in = (dleaf_t *) (mod_base + l->fileofs);

		if (l->filelen % sizeof (*in))
			Host_Error ("Mod_LoadLeafs: funny lump size in %s", loadmodel->name);

		count = l->filelen / sizeof (*in);
		out = (mleaf_t *) Hunk_AllocName (count * sizeof (*out), loadname);

		// johnfitz
		if (count > 32767)
			Host_Error ("Mod_LoadLeafs: %i leafs exceeds limit of 32767.\n", count);

		// johnfitz

		loadmodel->leafs = out;
		loadmodel->numleafs = count;

		for (i = 0; i < count; i++, in++, out++)
		{
			for (j = 0; j < 3; j++)
			{
				out->minmaxs[j] = LittleShort (in->mins[j]);
				out->minmaxs[3+j] = LittleShort (in->maxs[j]);
			}

			p = LittleLong (in->contents);
			out->contents = p;

			out->firstmarksurface = loadmodel->marksurfaces + (unsigned short) LittleShort (in->firstmarksurface); // johnfitz -- unsigned short
			out->nummarksurfaces = (unsigned short) LittleShort (in->nummarksurfaces); // johnfitz -- unsigned short

			p = LittleLong (in->visofs);

			if (p == -1)
				out->compressed_vis = NULL;
			else
				out->compressed_vis = loadmodel->visdata + p;

			out->efrags = NULL;

			for (j = 0; j < 4; j++)
				out->ambient_sound_level[j] = in->ambient_level[j];

			// johnfitz -- removed code to mark surfaces as SURF_UNDERWATER
		}
	}
}


/*
=================
Mod_LoadClipnodes
=================
*/
void Mod_SetBaseHulls (mclipnode_t *out, int count)
{
	hull_t *hull;

	// hull 0 is point-sized (0)
	hull = &loadmodel->hulls[1];
	hull->clipnodes = out;
	hull->firstclipnode = 0;
	hull->lastclipnode = count - 1;
	hull->planes = loadmodel->planes;
	hull->clip_mins[0] = -16;
	hull->clip_mins[1] = -16;
	hull->clip_mins[2] = -24;
	hull->clip_maxs[0] = 16;
	hull->clip_maxs[1] = 16;
	hull->clip_maxs[2] = 32;

	hull = &loadmodel->hulls[2];
	hull->clipnodes = out;
	hull->firstclipnode = 0;
	hull->lastclipnode = count - 1;
	hull->planes = loadmodel->planes;
	hull->clip_mins[0] = -32;
	hull->clip_mins[1] = -32;
	hull->clip_mins[2] = -24;
	hull->clip_maxs[0] = 32;
	hull->clip_maxs[1] = 32;
	hull->clip_maxs[2] = 64;
}


void Mod_LoadClipnodes (lump_t *l, int version)
{
	if (version == BSPVERSIONRMQ)
	{
		dclipnode29a_t *in;
		mclipnode_t *out; // johnfitz -- was dclipnode_t
		int			i, count;

		in = (dclipnode29a_t *) (mod_base + l->fileofs);

		if (l->filelen % sizeof (*in))
			Host_Error ("Mod_LoadClipnodes: funny lump size in %s", loadmodel->name);

		count = l->filelen / sizeof (*in);
		out = (mclipnode_t *) MallocZ (count * sizeof (*out));

		loadmodel->clipnodes = out;
		loadmodel->numclipnodes = count;

		Mod_SetBaseHulls (out, count);

		for (i = 0; i < count; i++, out++, in++)
		{
			out->planenum = LittleLong (in->planenum);

			// johnfitz -- bounds check
			if (out->planenum < 0 || out->planenum >= loadmodel->numplanes)
				Host_Error ("Mod_LoadClipnodes: planenum out of bounds");
			// johnfitz

			out->children[0] = LittleLong (in->children[0]);
			out->children[1] = LittleLong (in->children[1]);
		}
	}
	else
	{
		dclipnode_t *in;
		mclipnode_t *out; // johnfitz -- was dclipnode_t
		int			i, count;

		in = (dclipnode_t *) (mod_base + l->fileofs);

		if (l->filelen % sizeof (*in))
			Host_Error ("Mod_LoadClipnodes: funny lump size in %s", loadmodel->name);

		count = l->filelen / sizeof (*in);
		out = (mclipnode_t *) MallocZ (count * sizeof (*out));

		// johnfitz -- warn about exceeding old limits
		if (count > 32767) Con_DevWarning ("%i clipnodes exceeds standard limit of 32767.\n", count);
		// johnfitz

		loadmodel->clipnodes = out;
		loadmodel->numclipnodes = count;

		Mod_SetBaseHulls (out, count);

		for (i = 0; i < count; i++, out++, in++)
		{
			out->planenum = LittleLong (in->planenum);

			// johnfitz -- bounds check
			if (out->planenum < 0 || out->planenum >= loadmodel->numplanes)
				Host_Error ("Mod_LoadClipnodes: planenum out of bounds");
			// johnfitz

			// johnfitz -- support clipnodes > 32k
			out->children[0] = (unsigned short) LittleShort (in->children[0]);
			out->children[1] = (unsigned short) LittleShort (in->children[1]);

			if (out->children[0] >= count) out->children[0] -= 65536;
			if (out->children[1] >= count) out->children[1] -= 65536;

			if (out->children[0] < CONTENTS_CLIP) out->children[0] += 65536;
			if (out->children[1] < CONTENTS_CLIP) out->children[1] += 65536;
			// johnfitz
		}
	}
}


/*
=================
Mod_MakeHull0

Duplicate the drawing hull structure as a clipping hull
=================
*/
void Mod_MakeHull0 (void)
{
	mnode_t		*in, *child;
	mclipnode_t *out; // johnfitz -- was dclipnode_t
	int			i, j, count;
	hull_t		*hull;

	hull = &loadmodel->hulls[0];

	in = loadmodel->nodes;
	count = loadmodel->numnodes;
	out = (mclipnode_t *) MallocZ (count * sizeof (*out));

	hull->clipnodes = out;
	hull->firstclipnode = 0;
	hull->lastclipnode = count - 1;
	hull->planes = loadmodel->planes;

	for (i = 0; i < count; i++, out++, in++)
	{
		out->planenum = in->plane - loadmodel->planes;

		for (j = 0; j < 2; j++)
		{
			child = in->children[j];

			if (child->contents < 0)
				out->children[j] = child->contents;
			else out->children[j] = child - loadmodel->nodes;
		}
	}
}

/*
=================
Mod_LoadMarksurfaces
=================
*/
void Mod_LoadMarksurfaces (lump_t *l, int version)
{
	if (version == BSPVERSIONRMQ)
	{
		int		i, j, count;
		int		*in;
		msurface_t **out;

		in = (int *) (mod_base + l->fileofs);

		if (l->filelen % sizeof (*in))
			Host_Error ("Mod_LoadMarksurfaces: funny lump size in %s", loadmodel->name);

		count = l->filelen / sizeof (*in);
		out = (msurface_t **) MallocZ (count * sizeof (*out));

		loadmodel->marksurfaces = out;
		loadmodel->nummarksurfaces = count;

		for (i = 0; i < count; i++)
		{
			j = LittleLong (in[i]);

			if (j >= loadmodel->numsurfaces)
				Host_Error ("Mod_LoadMarksurfaces: bad surface number");

			out[i] = loadmodel->surfaces + j;
		}
	}
	else
	{
		int		i, j, count;
		short		*in;
		msurface_t **out;

		in = (short *) (mod_base + l->fileofs);

		if (l->filelen % sizeof (*in))
			Host_Error ("Mod_LoadMarksurfaces: funny lump size in %s", loadmodel->name);

		count = l->filelen / sizeof (*in);
		out = (msurface_t **) MallocZ (count * sizeof (*out));

		loadmodel->marksurfaces = out;
		loadmodel->nummarksurfaces = count;

		// johnfitz -- warn mappers about exceeding old limits
		if (count > 32767)
			Con_DevWarning ("%i marksurfaces exceeds standard limit of 32767.\n", count);
		// johnfitz

		for (i = 0; i < count; i++)
		{
			j = (unsigned short) LittleShort (in[i]); // johnfitz -- explicit cast as unsigned short

			if (j >= loadmodel->numsurfaces)
				Host_Error ("Mod_LoadMarksurfaces: bad surface number");

			out[i] = loadmodel->surfaces + j;
		}
	}
}

/*
=================
Mod_LoadSurfedges
=================
*/
void Mod_LoadSurfedges (lump_t *l, int version)
{
	int		i, count;
	int		*in, *out;

	in = (int *) (mod_base + l->fileofs);

	if (l->filelen % sizeof (*in))
		Host_Error ("Mod_LoadSurfedges: funny lump size in %s", loadmodel->name);

	count = l->filelen / sizeof (*in);
	out = (int *) MallocZ (count * sizeof (*out));

	loadmodel->surfedges = out;
	loadmodel->numsurfedges = count;

	for (i = 0; i < count; i++)
		out[i] = LittleLong (in[i]);
}


/*
=================
Mod_LoadPlanes
=================
*/
void Mod_LoadPlanes (lump_t *l, int version)
{
	int			i, j;
	mplane_t	*out;
	dplane_t 	*in;
	int			count;
	int			bits;

	in = (dplane_t *) (mod_base + l->fileofs);

	if (l->filelen % sizeof (*in))
		Host_Error ("Mod_LoadPlanes: funny lump size in %s", loadmodel->name);

	count = l->filelen / sizeof (*in);
	out = (mplane_t *) MallocZ (count * sizeof (*out));

	loadmodel->planes = out;
	loadmodel->numplanes = count;

	for (i = 0; i < count; i++, in++, out++)
	{
		bits = 0;

		for (j = 0; j < 3; j++)
		{
			out->normal[j] = LittleFloat (in->normal[j]);

			if (out->normal[j] < 0)
				bits |= 1 << j;
		}

		out->dist = LittleFloat (in->dist);
		out->type = LittleLong (in->type);
		out->signbits = bits;
	}
}

/*
=================
RadiusFromBounds
=================
*/
float RadiusFromBounds (vec3_t mins, vec3_t maxs)
{
	int		i;
	vec3_t	corner;

	for (i = 0; i < 3; i++)
	{
		corner[i] = fabs (mins[i]) > fabs (maxs[i]) ? fabs (mins[i]) : fabs (maxs[i]);
	}

	return Length (corner);
}

/*
=================
Mod_LoadSubmodels
=================
*/
void Mod_LoadSubmodels (lump_t *l, int version)
{
	dmodel_t	*in;
	dmodel_t	*out;
	int			i, j, count;

	in = (dmodel_t *) (mod_base + l->fileofs);

	if (l->filelen % sizeof (*in))
		Host_Error ("Mod_LoadSubmodels: funny lump size in %s", loadmodel->name);

	count = l->filelen / sizeof (*in);
	out = (dmodel_t *) MallocZ (count * sizeof (*out));

	loadmodel->submodels = out;
	loadmodel->numsubmodels = count;

	for (i = 0; i < count; i++, in++, out++)
	{
		for (j = 0; j < 3; j++)
		{
			// spread the mins / maxs by a pixel
			out->mins[j] = LittleFloat (in->mins[j]) - 1;
			out->maxs[j] = LittleFloat (in->maxs[j]) + 1;
			out->origin[j] = LittleFloat (in->origin[j]);
		}

		for (j = 0; j < MAX_MAP_HULLS; j++)
			out->headnode[j] = LittleLong (in->headnode[j]);

		out->visleafs = LittleLong (in->visleafs);
		out->firstface = LittleLong (in->firstface);
		out->numfaces = LittleLong (in->numfaces);
	}

	// johnfitz -- check world visleafs -- adapted from bjp
	out = loadmodel->submodels;

	if (out->visleafs > MAX_MAP_LEAFS)
		Host_Error ("Mod_LoadSubmodels: too many visleafs (%d, max = %d) in %s", out->visleafs, MAX_MAP_LEAFS, loadmodel->name);

	if (out->visleafs > 8192)
		Con_DevWarning ("%i visleafs exceeds standard limit of 8192.\n", out->visleafs);
	// johnfitz
}

/*
=================
Mod_BoundsFromClipNode -- johnfitz

update the model's clipmins and clipmaxs based on each node's plane.

This works because of the way brushes are expanded in hull generation.
Each brush will include all six axial planes, which bound that brush.
Therefore, the bounding box of the hull can be constructed entirely
from axial planes found in the clipnodes for that hull.
=================
*/
void Mod_BoundsFromClipNode (model_t *mod, int hull, int nodenum)
{
	mplane_t	*plane;
	mclipnode_t	*node;

	if (nodenum < 0)
		return; // hit a leafnode

	node = &mod->clipnodes[nodenum];
	plane = mod->hulls[hull].planes + node->planenum;

	switch (plane->type)
	{
	case PLANE_X:

		if (plane->signbits == 1)
			mod->clipmins[0] = min (mod->clipmins[0], -plane->dist - mod->hulls[hull].clip_mins[0]);
		else
			mod->clipmaxs[0] = max (mod->clipmaxs[0], plane->dist - mod->hulls[hull].clip_maxs[0]);

		break;
	case PLANE_Y:

		if (plane->signbits == 2)
			mod->clipmins[1] = min (mod->clipmins[1], -plane->dist - mod->hulls[hull].clip_mins[1]);
		else
			mod->clipmaxs[1] = max (mod->clipmaxs[1], plane->dist - mod->hulls[hull].clip_maxs[1]);

		break;
	case PLANE_Z:

		if (plane->signbits == 4)
			mod->clipmins[2] = min (mod->clipmins[2], -plane->dist - mod->hulls[hull].clip_mins[2]);
		else
			mod->clipmaxs[2] = max (mod->clipmaxs[2], plane->dist - mod->hulls[hull].clip_maxs[2]);

		break;
	default:
		// skip nonaxial planes; don't need them
		break;
	}

	Mod_BoundsFromClipNode (mod, hull, node->children[0]);
	Mod_BoundsFromClipNode (mod, hull, node->children[1]);
}


void Mod_SetCorrectBBox (model_t *mod)
{
	int i, j;
	msurface_t *surf;

	surf = mod->surfaces + mod->firstmodelsurface;
	mod->bbmins[0] = mod->bbmins[1] = mod->bbmins[2] = 9999999;
	mod->bbmaxs[0] = mod->bbmaxs[1] = mod->bbmaxs[2] = -9999999;

	for (i = 0; i < mod->nummodelsurfaces; i++, surf++)
	{
		for (j = 0; j < 3; j++)
		{
			if (surf->mins[j] < mod->bbmins[j]) mod->bbmins[j] = surf->mins[j];
			if (surf->maxs[j] > mod->bbmaxs[j]) mod->bbmaxs[j] = surf->maxs[j];
		}
	}
}


// split out to use common code for both BSP versions
void Mod_FinalizeBrushModel (model_t *mod, int version)
{
	int i, j;
	dmodel_t 	*bm;

	Mod_MakeHull0 ();

	mod->numframes = 2;		// regular and alternate animation

	// set up the submodels (FIXME: this is confusing)
	// johnfitz -- okay, so that i stop getting confused every time i look at this loop, here's how it works:
	// we're looping through the submodels starting at 0.  Submodel 0 is the main model, so we don't have to
	// worry about clobbering data the first time through, since it's the same data.  At the end of the loop,
	// we create a new copy of the data to use the next time through.
	for (i = 0; i < mod->numsubmodels; i++)
	{
		bm = &mod->submodels[i];

		mod->hulls[0].firstclipnode = bm->headnode[0];

		for (j = 1; j < MAX_MAP_HULLS; j++)
		{
			mod->hulls[j].firstclipnode = bm->headnode[j];
			mod->hulls[j].lastclipnode = mod->numclipnodes - 1;
		}

		mod->firstmodelsurface = bm->firstface;
		mod->nummodelsurfaces = bm->numfaces;

		// rmins/rmaxs and ymins/ymaxs were removed because they were kinda crap; we have *proper* rotating bbox support now ;)
		VectorCopy (bm->maxs, mod->maxs);
		VectorCopy (bm->mins, mod->mins);

		// johnfitz -- correct physics cullboxes so that outlying clip brushes on doors and stuff are handled right
		if (i > 0 || strcmp (mod->name, sv.modelname) != 0) // skip submodel 0 of sv.worldmodel, which is the actual world
		{
			// start with the hull0 bounds
			VectorCopy (mod->maxs, mod->clipmaxs);
			VectorCopy (mod->mins, mod->clipmins);

			// process hull1 (we don't need to process hull2 becuase there's
			// no such thing as a brush that appears in hull2 but not hull1)
			// Mod_BoundsFromClipNode (mod, 1, mod->hulls[1].firstclipnode); // (disabled for now becuase it fucks up on rotating models)
		}
		// johnfitz

		mod->numleafs = bm->visleafs;

		if (i < mod->numsubmodels - 1)
		{
			// duplicate the basic information
			char	name[10];

			sprintf (name, "*%i", i + 1);
			loadmodel = Mod_FindName (name);
			*loadmodel = *mod;
			strcpy (loadmodel->name, name);
			mod = loadmodel;
		}
	}
}


/*
=================
Mod_LoadBrushModel
=================
*/
void Mod_LoadBrushModel (model_t *mod, void *buffer)
{
	int			i, version;
	dheader_t	*header;

	loadmodel->type = mod_brush;

	header = (dheader_t *) buffer;

	version = LittleLong (header->version);

	if (version != BSPVERSION && version != BSPVERSIONRMQ)
		Host_Error ("Mod_LoadBrushModel: %s has wrong version number (%i should be %i or %i)", mod->name, version, BSPVERSION, BSPVERSIONRMQ);

	// swap all the lumps
	mod_base = (byte *) header;

	for (i = 0; i < sizeof (dheader_t) / 4; i++)
		((int *) header)[i] = LittleLong (((int *) header)[i]);

	// load into heap
	Mod_LoadVertexes (&header->lumps[LUMP_VERTEXES], version);
	Mod_LoadEdges (&header->lumps[LUMP_EDGES], version);
	Mod_LoadSurfedges (&header->lumps[LUMP_SURFEDGES], version);
	Mod_LoadTextures (&header->lumps[LUMP_TEXTURES], version);
	Mod_LoadLighting (&header->lumps[LUMP_LIGHTING], version);
	Mod_LoadPlanes (&header->lumps[LUMP_PLANES], version);
	Mod_LoadTexinfo (&header->lumps[LUMP_TEXINFO], version);
	Mod_LoadFaces (&header->lumps[LUMP_FACES], version);
	Mod_LoadMarksurfaces (&header->lumps[LUMP_MARKSURFACES], version);
	Mod_LoadVisibility (&header->lumps[LUMP_VISIBILITY], version);
	Mod_LoadLeafs (&header->lumps[LUMP_LEAFS], version);
	Mod_LoadNodes (&header->lumps[LUMP_NODES], version);
	Mod_LoadClipnodes (&header->lumps[LUMP_CLIPNODES], version);
	Mod_LoadEntities (&header->lumps[LUMP_ENTITIES], version);
	Mod_LoadSubmodels (&header->lumps[LUMP_MODELS], version);

	Mod_FinalizeBrushModel (mod, version);
}


/*
==============================================================================

IQM

==============================================================================
*/

void GL_AnimateIQMFrame (iqmdata_t *iqm, int currframe, int lastframe, float lerp);


void Mod_FixupIQMFlags (model_t *mod)
{
	// we don't have flags in IQMs so look for an original MDL and copy them out
	FILE *f = NULL;

	// mod->name is unchanged by the IQM loader (this is to preserve compatibility elsewhere in code)
	COM_FOpenFile (mod->name, &f);

	if (f)
	{
		mdl_t mdl;

		fread (&mdl, sizeof (mdl_t), 1, f);

		mdl.ident = LittleLong (mdl.ident);
		mdl.flags = LittleLong (mdl.flags);

		// make sure that it was an actual MDL
		if (mdl.ident == IDPOLYHEADER)
			mod->flags = mdl.flags;

		fclose (f);
	}
}


void Mod_ParseIQMModel (model_t *mod, iqmheader_t *iqm, char *path)
{
	int i, j;
	byte *buf = (byte *) iqm;
	char *str = (char *) &buf[iqm->ofs_text];
	iqmdata_t *iqmdata;
	iqmjointv1_t *jointv1;
	iqmjointv2_t *jointv2;
	matrix3x4_t *inversebaseframe;
	iqmposev1_t *posesv1;
	iqmposev2_t *posesv2;
	unsigned short *framedata;
	iqmbounds_t *bounds;
	int *inelements;
	iqmmesh_t *mesh;

	iqmvertexarray_t *va;
	float *vposition = NULL, *vtexcoord = NULL, *vnormal = NULL;

	// copy out flags
	mod->flags = iqm->flags;

	iqmdata = (iqmdata_t *) Hunk_Alloc (sizeof (iqmdata_t));

	va = (iqmvertexarray_t *) (buf + iqm->ofs_vertexarrays);

	for (i = 0; i < iqm->num_vertexarrays; i++)
	{
		va[i].type = LittleLong (va[i].type);
		va[i].flags = LittleLong (va[i].flags);
		va[i].format = LittleLong (va[i].format);
		va[i].size = LittleLong (va[i].size);
		va[i].offset = LittleLong (va[i].offset);

		switch (va[i].type)
		{
		case IQM_POSITION:
			if (va[i].format == IQM_FLOAT && va[i].size == 3)
				vposition = (float *) (buf + va[i].offset);
			break;

		case IQM_TEXCOORD:
			if (va[i].format == IQM_FLOAT && va[i].size == 2)
				vtexcoord = (float *) (buf + va[i].offset);
			break;

		case IQM_NORMAL:
			if (va[i].format == IQM_FLOAT && va[i].size == 3)
				vnormal = (float *) (buf + va[i].offset);
			break;

		case IQM_BLENDINDEXES:
			if (va[i].format == IQM_UBYTE && va[i].size == 4)
			{
				iqmdata->blendindexes = (unsigned char *) Hunk_Alloc (iqm->num_vertexes * 4 * sizeof (unsigned char));
				memcpy (iqmdata->blendindexes, (unsigned char *) (buf + va[i].offset), iqm->num_vertexes * 4 * sizeof (unsigned char));
			}
			break;

		case IQM_BLENDWEIGHTS:
			if (va[i].format == IQM_UBYTE && va[i].size == 4)
			{
				unsigned char *src;
				float *dst;

				iqmdata->blendweights = (float *) Hunk_Alloc (iqm->num_vertexes * 4 * sizeof (float));
				src = (unsigned char *) (buf + va[i].offset);
				dst = iqmdata->blendweights;

				for (j = 0; j < iqm->num_vertexes; j++, src += 4, dst += 4)
				{
					dst[0] = (float) src[0] / 255.0f;
					dst[1] = (float) src[1] / 255.0f;
					dst[2] = (float) src[2] / 255.0f;
					dst[3] = (float) src[3] / 255.0f;
				}
			}
			break;
		}
	}

	if (!vposition || !vtexcoord || !vnormal || !iqmdata->blendindexes || !iqmdata->blendweights)
		return;

	iqmdata->num_frames = iqm->num_anims;
	iqmdata->num_joints = iqm->num_joints;
	iqmdata->num_poses = iqm->num_frames;
	iqmdata->numvertexes = iqm->num_vertexes;
	iqmdata->num_triangles = iqm->num_triangles;

	if (iqm->version == IQM_VERSION1)
	{
		// load the joints
		jointv1 = (iqmjointv1_t *) (buf + iqm->ofs_joints);
		iqmdata->jointsv1 = (iqmjointv1_t *) Hunk_Alloc (iqm->num_joints * sizeof (iqmjointv1_t));

		for (i = 0; i < iqmdata->num_joints; i++)
		{
			iqmdata->jointsv1[i].name = LittleLong (jointv1[i].name);
			iqmdata->jointsv1[i].parent = LittleLong (jointv1[i].parent);

			for (j = 0; j < 3; j++)
			{
				iqmdata->jointsv1[i].translate[j] = LittleFloat (jointv1[i].translate[j]);
				iqmdata->jointsv1[i].rotate[j] = LittleFloat (jointv1[i].rotate[j]);
				iqmdata->jointsv1[i].scale[j] = LittleFloat (jointv1[i].scale[j]);
			}
		}
	}
	else
	{
		// load the joints
		jointv2 = (iqmjointv2_t *) (buf + iqm->ofs_joints);
		iqmdata->jointsv2 = (iqmjointv2_t *) Hunk_Alloc (iqm->num_joints * sizeof (iqmjointv2_t));

		for (i = 0; i < iqmdata->num_joints; i++)
		{
			iqmdata->jointsv2[i].name = LittleLong (jointv2[i].name);
			iqmdata->jointsv2[i].parent = LittleLong (jointv2[i].parent);

			for (j = 0; j < 3; j++)
			{
				iqmdata->jointsv2[i].translate[j] = LittleFloat (jointv2[i].translate[j]);
				iqmdata->jointsv2[i].rotate[j] = LittleFloat (jointv2[i].rotate[j]);
				iqmdata->jointsv2[i].scale[j] = LittleFloat (jointv2[i].scale[j]);
			}

			iqmdata->jointsv2[i].rotate[3] = LittleFloat (jointv2[i].rotate[3]);
		}
	}

	// these don't need to be a part of mod - remember to free them
	iqmdata->baseframe = (matrix3x4_t *) Hunk_Alloc (iqm->num_joints * sizeof (matrix3x4_t));
	inversebaseframe = (matrix3x4_t *) malloc (iqm->num_joints * sizeof (matrix3x4_t));

	if (iqm->version == IQM_VERSION1)
	{
		for (i = 0; i < (int) iqm->num_joints; i++)
		{
			vec3_t rot;
			vec4_t q_rot;
			iqmjointv1_t j = jointv1[i];

			// first need to make a vec4 quat from our rotation vec
			VectorSet (rot, j.rotate[0], j.rotate[1], j.rotate[2]);
			Vector4Set (q_rot, j.rotate[0], j.rotate[1], j.rotate[2], -sqrt (max (1.0 - pow (VectorLength (rot), 2), 0.0)));

			Matrix3x4_FromQuatAndVectors (&iqmdata->baseframe[i], q_rot, j.translate, j.scale);
			Matrix3x4_Invert (&inversebaseframe[i], iqmdata->baseframe[i]);

			if (j.parent >= 0)
			{
				matrix3x4_t temp;
				Matrix3x4_Multiply (&temp, iqmdata->baseframe[j.parent], iqmdata->baseframe[i]);
				iqmdata->baseframe[i] = temp;
				Matrix3x4_Multiply (&temp, inversebaseframe[i], inversebaseframe[j.parent]);
				inversebaseframe[i] = temp;
			}
		}
	}
	else
	{
		for (i = 0; i < (int) iqm->num_joints; i++)
		{
			iqmjointv2_t j = jointv2[i];

			Matrix3x4_FromQuatAndVectors (&iqmdata->baseframe[i], j.rotate, j.translate, j.scale);
			Matrix3x4_Invert (&inversebaseframe[i], iqmdata->baseframe[i]);

			if (j.parent >= 0)
			{
				matrix3x4_t temp;
				Matrix3x4_Multiply (&temp, iqmdata->baseframe[j.parent], iqmdata->baseframe[i]);
				iqmdata->baseframe[i] = temp;
				Matrix3x4_Multiply (&temp, inversebaseframe[i], inversebaseframe[j.parent]);
				inversebaseframe[i] = temp;
			}
		}
	}

	if (iqm->version == IQM_VERSION1)
	{
		posesv1 = (iqmposev1_t *) (buf + iqm->ofs_poses);
		iqmdata->frames = (matrix3x4_t *) Hunk_Alloc (iqm->num_frames * iqm->num_poses * sizeof (matrix3x4_t));
		framedata = (unsigned short *) (buf + iqm->ofs_frames);

		for (i = 0; i < iqm->num_frames; i++)
		{
			for (j = 0; j < iqm->num_poses; j++)
			{
				iqmposev1_t p = posesv1[j];
				vec3_t translate, rotate, scale;
				vec4_t q_rot;
				matrix3x4_t m, temp;

				translate[0] = p.channeloffset[0]; if (p.mask & 0x01) translate[0] += *framedata++ * p.channelscale[0];
				translate[1] = p.channeloffset[1]; if (p.mask & 0x02) translate[1] += *framedata++ * p.channelscale[1];
				translate[2] = p.channeloffset[2]; if (p.mask & 0x04) translate[2] += *framedata++ * p.channelscale[2];
				rotate[0] = p.channeloffset[3]; if (p.mask & 0x08) rotate[0] += *framedata++ * p.channelscale[3];
				rotate[1] = p.channeloffset[4]; if (p.mask & 0x10) rotate[1] += *framedata++ * p.channelscale[4];
				rotate[2] = p.channeloffset[5]; if (p.mask & 0x20) rotate[2] += *framedata++ * p.channelscale[5];
				scale[0] = p.channeloffset[6]; if (p.mask & 0x40) scale[0] += *framedata++ * p.channelscale[6];
				scale[1] = p.channeloffset[7]; if (p.mask & 0x80) scale[1] += *framedata++ * p.channelscale[7];
				scale[2] = p.channeloffset[8]; if (p.mask & 0x100) scale[2] += *framedata++ * p.channelscale[8];
				// Concatenate each pose with the inverse base pose to avoid doing this at animation time.
				// If the joint has a parent, then it needs to be pre-concatenated with its parent's base pose.
				// Thus it all negates at animation time like so:
				//   (parentPose * parentInverseBasePose) * (parentBasePose * childPose * childInverseBasePose) =>
				//   parentPose * (parentInverseBasePose * parentBasePose) * childPose * childInverseBasePose =>
				//   parentPose * childPose * childInverseBasePose

				Vector4Set (q_rot, rotate[0], rotate[1], rotate[2], -sqrt (max (1.0 - pow (VectorLength (rotate), 2), 0.0)));
				Matrix3x4_FromQuatAndVectors (&m, q_rot, translate, scale);

				if (p.parent >= 0)
				{
					Matrix3x4_Multiply (&temp, iqmdata->baseframe[p.parent], m);
					Matrix3x4_Multiply (&iqmdata->frames[i * iqm->num_poses + j], temp, inversebaseframe[j]);
				}
				else Matrix3x4_Multiply (&iqmdata->frames[i * iqm->num_poses + j], m, inversebaseframe[j]);
			}
		}
	}
	else
	{
		posesv2 = (iqmposev2_t *) (buf + iqm->ofs_poses);
		iqmdata->frames = (matrix3x4_t *) Hunk_Alloc (iqm->num_frames * iqm->num_poses * sizeof (matrix3x4_t));
		framedata = (unsigned short *) (buf + iqm->ofs_frames);

		for (i = 0; i < iqm->num_frames; i++)
		{
			for (j = 0; j < iqm->num_poses; j++)
			{
				iqmposev2_t p = posesv2[j];
				vec3_t translate, scale;
				vec4_t rotate;
				matrix3x4_t m, temp;

				translate[0] = p.channeloffset[0]; if (p.mask & 0x01) translate[0] += *framedata++ * p.channelscale[0];
				translate[1] = p.channeloffset[1]; if (p.mask & 0x02) translate[1] += *framedata++ * p.channelscale[1];
				translate[2] = p.channeloffset[2]; if (p.mask & 0x04) translate[2] += *framedata++ * p.channelscale[2];
				rotate[0] = p.channeloffset[3]; if (p.mask & 0x08) rotate[0] += *framedata++ * p.channelscale[3];
				rotate[1] = p.channeloffset[4]; if (p.mask & 0x10) rotate[1] += *framedata++ * p.channelscale[4];
				rotate[2] = p.channeloffset[5]; if (p.mask & 0x20) rotate[2] += *framedata++ * p.channelscale[5];
				rotate[3] = p.channeloffset[6]; if (p.mask & 0x40) rotate[3] += *framedata++ * p.channelscale[6];
				scale[0] = p.channeloffset[7]; if (p.mask & 0x80) scale[0] += *framedata++ * p.channelscale[7];
				scale[1] = p.channeloffset[8]; if (p.mask & 0x100) scale[1] += *framedata++ * p.channelscale[8];
				scale[2] = p.channeloffset[9]; if (p.mask & 0x200) scale[2] += *framedata++ * p.channelscale[9];
				// Concatenate each pose with the inverse base pose to avoid doing this at animation time.
				// If the joint has a parent, then it needs to be pre-concatenated with its parent's base pose.
				// Thus it all negates at animation time like so:
				//   (parentPose * parentInverseBasePose) * (parentBasePose * childPose * childInverseBasePose) =>
				//   parentPose * (parentInverseBasePose * parentBasePose) * childPose * childInverseBasePose =>
				//   parentPose * childPose * childInverseBasePose

				Matrix3x4_FromQuatAndVectors (&m, rotate, translate, scale);

				if (p.parent >= 0)
				{
					Matrix3x4_Multiply (&temp, iqmdata->baseframe[p.parent], m);
					Matrix3x4_Multiply (&iqmdata->frames[i * iqm->num_poses + j], temp, inversebaseframe[j]);
				}
				else Matrix3x4_Multiply (&iqmdata->frames[i * iqm->num_poses + j], m, inversebaseframe[j]);
			}
		}
	}

	iqmdata->outframe = (matrix3x4_t *) Hunk_Alloc (iqmdata->num_joints * sizeof (matrix3x4_t));

	// load bounding box data
	if (iqm->ofs_bounds)
	{
		float xyradius = 0, radius = 0;

		bounds = (iqmbounds_t *) (buf + iqm->ofs_bounds);

		// vectorclear seems dubious as either could go on the wrong side of 0
		iqmdata->mins[0] = 99999999;
		iqmdata->mins[1] = 99999999;
		iqmdata->mins[2] = 99999999;

		iqmdata->maxs[0] = -99999999;
		iqmdata->maxs[1] = -99999999;
		iqmdata->maxs[2] = -99999999;

		// we're only using one frame so we only use one bounding box for the entire model
		// we'll still check 'em all just to be precise though
		for (i = 0; i < (int) iqm->num_frames; i++)
		{
			for (j = 0; j < 3; j++)
			{
				bounds[i].bbmin[j] = LittleFloat (bounds[i].bbmin[j]);
				bounds[i].bbmax[j] = LittleFloat (bounds[i].bbmax[j]);

				if (bounds[i].bbmin[j] < iqmdata->mins[j]) iqmdata->mins[j] = bounds[i].bbmin[j];
				if (bounds[i].bbmax[j] > iqmdata->maxs[j]) iqmdata->maxs[j] = bounds[i].bbmax[j];
			}

			bounds[i].xyradius = LittleFloat (bounds[i].xyradius);
			bounds[i].radius = LittleFloat (bounds[i].radius);

			if (bounds[i].xyradius > xyradius) xyradius = bounds[i].xyradius;
			if (bounds[i].radius > radius) radius = bounds[i].radius;
		}

		for (i = 0; i < 3; i++)
		{
			mod->mins[i] = iqmdata->mins[i];
			mod->maxs[i] = iqmdata->maxs[i];
		}

		iqmdata->radius = radius;
	}

	// load triangle data
	inelements = (int *) (buf + iqm->ofs_triangles);
	iqmdata->tris = (iqmtriangle_t *) Hunk_Alloc (iqm->num_triangles * sizeof (iqmtriangle_t));

	for (i = 0; i < (int) iqm->num_triangles; i++)
	{
		iqmdata->tris[i].vertex[0] = LittleLong (inelements[0]);
		iqmdata->tris[i].vertex[1] = LittleLong (inelements[1]);
		iqmdata->tris[i].vertex[2] = LittleLong (inelements[2]);
		inelements += 3;
	}

	// lead meshes
	iqmdata->num_mesh = iqm->num_meshes;
	iqmdata->mesh = (iqmmesh_t *) Hunk_Alloc (iqmdata->num_mesh * sizeof (iqmmesh_t));
	iqmdata->indexoffsets = (int *) Hunk_Alloc (iqmdata->num_mesh * sizeof (int));	// only vbos need this but it's a handful of ints so blah
	iqmdata->skins = (gltexture_t **) Hunk_Alloc (iqmdata->num_mesh * sizeof (gltexture_t *));
	iqmdata->fullbrights = (gltexture_t **) Hunk_Alloc (iqmdata->num_mesh * sizeof (gltexture_t *));
	mesh = (iqmmesh_t *) (buf + iqm->ofs_meshes);

	for (i = 0; i < iqmdata->num_mesh; i++)
	{
		char texturepath[256];
		byte *data = NULL;
		int w, h;

		// Image_LoadImage puts data on the hunk and it's our resposibility to free it
		int hunkmark = Hunk_LowMark ();

		iqmdata->mesh[i].first_triangle = LittleLong (mesh[i].first_triangle);
		iqmdata->mesh[i].first_vertex = LittleLong (mesh[i].first_vertex);
		iqmdata->mesh[i].material = LittleLong (mesh[i].material);
		iqmdata->mesh[i].name = LittleLong (mesh[i].name);
		iqmdata->mesh[i].num_vertexes = LittleLong (mesh[i].num_vertexes);
		iqmdata->mesh[i].num_triangles = LittleLong (mesh[i].num_triangles);

		// build the base texture path from the model directory and the material defined in the IQM
		strcpy (texturepath, path);

		for (j = strlen (texturepath); j; j--)
		{
			if (texturepath[j] == '/' || texturepath[j] == '\\')
			{
				strcpy (&texturepath[j + 1], &str[iqmdata->mesh[i].material]);
				break;
			}
		}

		// rip the extension for Fitz's texture loader
		for (j = strlen (texturepath); j; j--)
		{
			if (texturepath[j] == '.')
			{
				texturepath[j] = 0;
				break;
			}
		}

		// load regular skin
		data = Image_LoadImage (texturepath, &w, &h);

		if (data)
			iqmdata->skins[i] = TM_LoadImage (mod, texturepath, w, h, SRC_RGBA, data, texturepath, 0, TEXPREF_MIPMAP);
		else iqmdata->skins[i] = NULL;

		strcat (texturepath, "_luma");

		// load optional fullbright
		data = Image_LoadImage (texturepath, &w, &h);

		if (data)
			iqmdata->fullbrights[i] = TM_LoadImage (mod, texturepath, w, h, SRC_RGBA, data, texturepath, 0, TEXPREF_MIPMAP);
		else iqmdata->fullbrights[i] = NULL;

		Con_DPrintf ("texture %i %s\n", i, texturepath);

		Hunk_FreeToLowMark (hunkmark);
	}

	// load vertex data
	iqmdata->verts = (iqmvertex_t *) Hunk_Alloc (iqmdata->numvertexes * sizeof (iqmvertex_t));

	for (i = 0; i < iqmdata->numvertexes; i++)
	{
		VectorSet (iqmdata->verts[i].position, LittleFloat (vposition[0]), LittleFloat (vposition[1]), LittleFloat (vposition[2]));
		VectorSet (iqmdata->verts[i].normal, LittleFloat (vnormal[0]), LittleFloat (vnormal[1]), LittleFloat (vnormal[2]));

		vposition += 3;
		vnormal += 3;
	}

	for (i = 0; i < (int) iqm->num_vertexes; i++)
	{
		iqmdata->verts[i].texcoord[0] = LittleFloat (vtexcoord[0]);
		iqmdata->verts[i].texcoord[1] = LittleFloat (vtexcoord[1]);

		vtexcoord += 2;
	}

	// store the version so that we know how to animate it
	iqmdata->version = iqm->version;

	// if there is only one pose (e.g. in a barrel, crate, ammo box, etc) then pre-animate the model for faster loads
	// this will also update a valid cache and ensure that the animation is one-time-only (here)
	if (iqmdata->num_poses == 1) GL_AnimateIQMFrame (iqmdata, 0, 0, 1);

	// free temp non hunk mem
	if (inversebaseframe)
		free (inversebaseframe);

	// set up an invalidated cache to force it to animate first time it's seen
	iqmdata->cachedcurrframe = -1;
	iqmdata->cachedlastframe = -1;
	iqmdata->cachedlerp = 666;

	// loaded OK
	mod->iqmdata = iqmdata;
	mod->type = mod_iqm;

	if (!iqm->flags)
		Mod_FixupIQMFlags (mod);
	else mod->flags = iqm->flags;
} 


void Mod_SwapIQMHeader (iqmheader_t *header)
{
	// byteswap header
	header->version = LittleLong (header->version);
	header->filesize = LittleLong (header->filesize);
	header->flags = LittleLong (header->flags);
	header->num_text = LittleLong (header->num_text);
	header->ofs_text = LittleLong (header->ofs_text);
	header->num_meshes = LittleLong (header->num_meshes);
	header->ofs_meshes = LittleLong (header->ofs_meshes);
	header->num_vertexarrays = LittleLong (header->num_vertexarrays);
	header->num_vertexes = LittleLong (header->num_vertexes);
	header->ofs_vertexarrays = LittleLong (header->ofs_vertexarrays);
	header->num_triangles = LittleLong (header->num_triangles);
	header->ofs_triangles = LittleLong (header->ofs_triangles);
	header->ofs_adjacency = LittleLong (header->ofs_adjacency);
	header->num_joints = LittleLong (header->num_joints);
	header->ofs_joints = LittleLong (header->ofs_joints);
	header->num_poses = LittleLong (header->num_poses);
	header->ofs_poses = LittleLong (header->ofs_poses);
	header->num_anims = LittleLong (header->num_anims);
	header->ofs_anims = LittleLong (header->ofs_anims);
	header->num_frames = LittleLong (header->num_frames);
	header->num_framechannels = LittleLong (header->num_framechannels);
	header->ofs_frames = LittleLong (header->ofs_frames);
	header->num_comment = LittleLong (header->num_comment);
	header->ofs_comment = LittleLong (header->ofs_comment);
	header->num_extensions = LittleLong (header->num_extensions);
	header->ofs_extensions = LittleLong (header->ofs_extensions);
}


void Mod_LoadIQMModel (model_t *mod, void *buffer, char *path, qboolean crash)
{
	iqmheader_t *hdr = (iqmheader_t *) buffer;

	Mod_SwapIQMHeader (hdr);

	if (strcmp (hdr->magic, IQM_MAGIC)) goto IQM_LoadError;
	if (hdr->version != IQM_VERSION1 && hdr->version != IQM_VERSION2) goto IQM_LoadError;

	// ???????
	memcpy (hdr, buffer, hdr->filesize);

	Mod_ParseIQMModel (mod, hdr, path);
	return;

IQM_LoadError:;
	if (crash) Sys_Error ("Mod_LoadIQMModel : failed to load %s\n", loadname);
}


qboolean Mod_FindIQMModel (model_t *mod)
{
	FILE *f = NULL;
	char iqmname[MAX_OSPATH];

	// we just need to set this to something that's not mod_iqm so that we can identify if it loaded OK later on
	mod->type = mod_alias;

	// attempt q2 naming convention
	COM_StripExtension (mod->name, iqmname);
	strcat (iqmname, "/tris.iqm");

	COM_FOpenFile (iqmname, &f);

	if (f && com_filesize > 0)
	{
		iqmheader_t *hdr = (iqmheader_t *) malloc (com_filesize);

		fread (hdr, com_filesize, 1, f);
		fclose (f);

		Mod_LoadIQMModel (mod, hdr, iqmname, false);

		free (hdr);

		// and did it load?
		if (mod->type == mod_iqm) return true;
	}

	COM_StripExtension (mod->name, iqmname);
	COM_DefaultExtension (iqmname, ".iqm");

	COM_FOpenFile (iqmname, &f);

	if (f && com_filesize > 0)
	{
		iqmheader_t *hdr = (iqmheader_t *) malloc (com_filesize);

		fread (hdr, com_filesize, 1, f);
		fclose (f);

		Mod_LoadIQMModel (mod, hdr, iqmname, false);

		free (hdr);

		// and did it load?
		if (mod->type == mod_iqm) return true;
	}

	return false;
}


/*
==============================================================================

ALIAS MODELS

==============================================================================
*/


aliashdr_t *pheader;

void Mod_LoadAliasBBoxes (void)
{
	int i, n, v;
	aliasbbox_t *bboxes;
	trivertx_t *verts;

	bboxes = (aliasbbox_t *) Hunk_Alloc (pheader->nummeshframes * sizeof (aliasbbox_t));
	pheader->bboxes = (intptr_t) bboxes - (intptr_t) pheader;

	for (i = 0; i < pheader->nummeshframes; i++, bboxes++)
	{
		bboxes->mins[0] = bboxes->mins[1] = bboxes->mins[2] = 9999999;
		bboxes->maxs[0] = bboxes->maxs[1] = bboxes->maxs[2] = -9999999;

		verts = (trivertx_t *) ((byte *) pheader + pheader->vertexes + pheader->framevertexsize * i);

		for (n = 0; n < pheader->vertsperframe; n++, verts++)
		{
			for (v = 0; v < 3; v++)
			{
				if (verts->v[v] < bboxes->mins[v]) bboxes->mins[v] = verts->v[v];
				if (verts->v[v] > bboxes->maxs[v]) bboxes->maxs[v] = verts->v[v];
			}
		}

		for (n = 0; n < 3; n++)
		{
			bboxes->mins[n] = (bboxes->mins[n] * pheader->scale[n]) + pheader->scale_origin[n];
			bboxes->maxs[n] = (bboxes->maxs[n] * pheader->scale[n]) + pheader->scale_origin[n];
		}
	}
}


void Mod_LoadFrameVerts (maliasframedesc_t *frame, trivertx_t *verts)
{
	int i, v;
	int mark = Hunk_LowMark ();
	trivertx_t *vertexes = (trivertx_t *) Hunk_Alloc (pheader->vertsperframe * sizeof (trivertx_t));

	// this should only be done once and stores an offset to the first set of verts for the frame
	if (!pheader->vertexes) pheader->vertexes = (intptr_t) vertexes - (intptr_t) pheader;

	for (i = 0; i < pheader->vertsperframe; i++, vertexes++, verts++)
	{
		vertexes->lightnormalindex = verts->lightnormalindex;

		for (v = 0; v < 3; v++)
		{
			if (verts->v[v] < loadmodel->mins[v]) loadmodel->mins[v] = verts->v[v];
			if (verts->v[v] > loadmodel->maxs[v]) loadmodel->maxs[v] = verts->v[v];

			vertexes->v[v] = verts->v[v];
		}
	}

	// hunk tags will mean that pheader->vertsperframe * sizeof (trivertx_t) is invalid for this
	// we only need to do this once but no harm in doing it every frame
	pheader->framevertexsize = Hunk_LowMark () - mark;

	pheader->nummeshframes++;
}


void *Mod_LoadAliasFrame (void *pin, maliasframedesc_t *frame)
{
	trivertx_t		*verts;
	daliasframe_t	*pdaliasframe;

	pdaliasframe = (daliasframe_t *) pin;

	strncpy (frame->name, pdaliasframe->name, 16);
	frame->firstpose = pheader->nummeshframes;
	frame->numposes = 1;

	verts = (trivertx_t *) (pdaliasframe + 1);

	// load the frame vertexes
	Mod_LoadFrameVerts (frame, verts);
	verts += pheader->vertsperframe;

	return (void *) verts;
}


void *Mod_LoadAliasGroup (void *pin,  maliasframedesc_t *frame)
{
	daliasgroup_t		*pingroup;
	int					i, numframes;
	daliasinterval_t	*pin_intervals;
	void				*ptemp;

	pingroup = (daliasgroup_t *) pin;
	numframes = LittleLong (pingroup->numframes);
	frame->firstpose = pheader->nummeshframes;
	frame->numposes = numframes;

	pin_intervals = (daliasinterval_t *) (pingroup + 1);
	frame->interval = LittleFloat (pin_intervals->interval);
	pin_intervals += numframes;
	ptemp = (void *) pin_intervals;

	for (i = 0; i < numframes; i++)
	{
		Mod_LoadFrameVerts (frame, (trivertx_t *) ((daliasframe_t *) ptemp + 1));
		ptemp = (trivertx_t *) ((daliasframe_t *) ptemp + 1) + pheader->vertsperframe;
	}

	return ptemp;
}



/*
=================
Mod_FloodFillSkin

Fill background pixels so mipmapping doesn't have haloes - Ed
=================
*/

typedef struct
{
	short		x, y;
} floodfill_t;

extern unsigned d_8to24table[];

// must be a power of 2
#define FLOODFILL_FIFO_SIZE 0x1000
#define FLOODFILL_FIFO_MASK (FLOODFILL_FIFO_SIZE - 1)

#define FLOODFILL_STEP( off, dx, dy ) \
{ \
	if (pos[off] == fillcolor) \
	{ \
		pos[off] = 255; \
		fifo[inpt].x = x + (dx), fifo[inpt].y = y + (dy); \
		inpt = (inpt + 1) & FLOODFILL_FIFO_MASK; \
	} \
	else if (pos[off] != 255) fdc = pos[off]; \
}

void Mod_FloodFillSkin (byte *skin, int skinwidth, int skinheight)
{
	byte				fillcolor = *skin; // assume this is the pixel to fill
	// floodfill_t			fifo[FLOODFILL_FIFO_SIZE];
	int					inpt = 0, outpt = 0;
	int					filledcolor = -1;
	int					i;
	floodfill_t			*fifo = (floodfill_t *) scratchbuf;

	if (filledcolor == -1)
	{
		filledcolor = 0;

		// attempt to find opaque black
		for (i = 0; i < 256; ++i)
		{
			if (d_8to24table[i] == (255 << 0)) // alpha 1.0
			{
				filledcolor = i;
				break;
			}
		}
	}

	// can't fill to filled color or to transparent color (used as visited marker)
	if ((fillcolor == filledcolor) || (fillcolor == 255))
	{
		// printf( "not filling skin from %d to %d\n", fillcolor, filledcolor );
		return;
	}

	fifo[inpt].x = 0, fifo[inpt].y = 0;
	inpt = (inpt + 1) & FLOODFILL_FIFO_MASK;

	while (outpt != inpt)
	{
		int			x = fifo[outpt].x, y = fifo[outpt].y;
		int			fdc = filledcolor;
		byte		*pos = &skin[x + skinwidth * y];

		outpt = (outpt + 1) & FLOODFILL_FIFO_MASK;

		if (x > 0)				FLOODFILL_STEP (-1, -1, 0);
		if (x < skinwidth - 1)	FLOODFILL_STEP (1, 1, 0);
		if (y > 0)				FLOODFILL_STEP (-skinwidth, 0, -1);
		if (y < skinheight - 1)	FLOODFILL_STEP (skinwidth, 0, 1);

		skin[x + skinwidth *y] = fdc;
	}
}


#define MDL_STANDARD (TEXPREF_MIPMAP | TEXPREF_PAD | TEXPREF_FLOODFILL)
#define MDL_NOBRIGHT (TEXPREF_MIPMAP | TEXPREF_PAD | TEXPREF_NOBRIGHT | TEXPREF_FLOODFILL)
#define MDL_FULLBRIGHT	(TEXPREF_MIPMAP | TEXPREF_PAD | TEXPREF_FULLBRIGHT | TEXPREF_FLOODFILL)

char *texpaths[] = {"textures/progs/", "textures/models/", "progs/", "models/", NULL};

void Mod_LoadSkinTexture (gltexture_t **tex, gltexture_t **fb, char *name, char *fbr_mask_name, char *extname, byte *data, int width, int height, src_offset_t offset)
{
	int i;
	char extload[256] = {0};
	int extwidth = 0;
	int extheight = 0;
	byte *extdata;
	int hunkmark = Hunk_LowMark ();

	for (i = 0; ; i++)
	{
		// no more paths
		if (!texpaths[i]) break;

		sprintf (extload, "%s%s", texpaths[i], extname);
		extdata = Image_LoadImage (extload, &extwidth, &extheight);

		if (extdata)
		{
			tex[0] = TM_LoadImage (loadmodel, extload, extwidth, extheight, SRC_RGBA, extdata, extload, 0, TEXPREF_MIPMAP);
			fb[0] = NULL;

			// now try to load glow/luma image from the same place
			Hunk_FreeToLowMark (hunkmark);

			sprintf (extload, "%s%s_glow", texpaths[i], extname);
			extdata = Image_LoadImage (extload, &extwidth, &extheight);

			if (!extdata)
			{
				sprintf (extload, "%s%s_luma", texpaths[i], extname);
				extdata = Image_LoadImage (extload, &extwidth, &extheight);
			}

			if (extdata)
			{
				fb[0] = TM_LoadImage (loadmodel, extload, extwidth, extheight, SRC_RGBA, extdata, extload, 0, TEXPREF_MIPMAP);
				Hunk_FreeToLowMark (hunkmark);
			}

			return;
		}
	}

	// hurrah! for readable code
	if (Mod_CheckFullbrights (data, width * height))
	{
		tex[0] = TM_LoadImage (loadmodel, name, width, height, SRC_INDEXED, data, loadmodel->name, offset, MDL_NOBRIGHT);
		fb[0] = TM_LoadImage (loadmodel, fbr_mask_name, width, height, SRC_INDEXED, data, loadmodel->name, offset, MDL_FULLBRIGHT);
	}
	else
	{
		tex[0] = TM_LoadImage (loadmodel, name, width, height, SRC_INDEXED, data, loadmodel->name, offset, MDL_STANDARD);
		fb[0] = NULL;
	}

	// just making sure...
	Hunk_FreeToLowMark (hunkmark);
}


/*
===============
Mod_LoadAllSkins
===============
*/
void *Mod_LoadAllSkins (int numskins, daliasskintype_t *pskintype)
{
	int						i, j, k, size, groupskins;
	char					name[256];	// 32 this isn;t long enough for some mods
	byte					*skin, *texels;
	daliasskingroup_t		*pinskingroup;
	daliasskininterval_t	*pinskinintervals;
	char					fbr_mask_name[256]; // johnfitz -- added for fullbright support
	char					extname[256];
	src_offset_t				offset; // johnfitz

	skin = (byte *) (pskintype + 1);

	if (numskins < 1 || numskins > MAX_SKINS)
		Host_Error ("Mod_LoadAliasModel: Invalid # of skins: %d\n", numskins);

	size = pheader->skinwidth * pheader->skinheight;

	for (i = 0; i < numskins; i++)
	{
		if (pskintype->type == ALIAS_SKIN_SINGLE)
		{
			Mod_FloodFillSkin (skin, pheader->skinwidth, pheader->skinheight);

			// save 8 bit texels for the player model to remap
			texels = (byte *) Hunk_AllocName (size, loadname);
			pheader->texels[i] = texels - (byte *) pheader;
			memcpy (texels, (byte *) (pskintype + 1), size);

			// johnfitz -- rewritten
			sprintf (name, "%s:frame%i", loadmodel->name, i);
			sprintf (extname, "%s_%i", loadmodel->name, i);
			sprintf (fbr_mask_name, "%s:frame%i_glow", loadmodel->name, i);
			offset = (src_offset_t) (pskintype + 1) - (src_offset_t) mod_base;

			Mod_LoadSkinTexture
			(
				&pheader->gltextures[i][0],
				&pheader->fbtextures[i][0],
				name,
				fbr_mask_name,
				&extname[6],
				(byte *) (pskintype + 1),
				pheader->skinwidth,
				pheader->skinheight,
				offset
			);

			pheader->gltextures[i][3] = pheader->gltextures[i][2] = pheader->gltextures[i][1] = pheader->gltextures[i][0];
			pheader->fbtextures[i][3] = pheader->fbtextures[i][2] = pheader->fbtextures[i][1] = pheader->fbtextures[i][0];
			// johnfitz

			pskintype = (daliasskintype_t *) ((byte *) (pskintype + 1) + size);
		}
		else
		{
			// animating skin group.  yuck.
			pskintype++;
			pinskingroup = (daliasskingroup_t *) pskintype;
			groupskins = LittleLong (pinskingroup->numskins);
			pinskinintervals = (daliasskininterval_t *) (pinskingroup + 1);

			pskintype = (daliasskintype_t *) (pinskinintervals + groupskins);

			for (j = 0; j < groupskins; j++)
			{
				Mod_FloodFillSkin (skin, pheader->skinwidth, pheader->skinheight);

				if (j == 0)
				{
					texels = (byte *) Hunk_AllocName (size, loadname);
					pheader->texels[i] = texels - (byte *) pheader;
					memcpy (texels, (byte *) (pskintype), size);
				}

				// johnfitz -- rewritten
				sprintf (name, "%s:frame%i_%i", loadmodel->name, i, j);
				sprintf (extname, "%s_%i_%i", loadmodel->name, i, j);
				sprintf (fbr_mask_name, "%s:frame%i_%i_glow", loadmodel->name, i, j);
				offset = (src_offset_t) (pskintype) - (src_offset_t) mod_base; // johnfitz

				Mod_LoadSkinTexture
				(
					&pheader->gltextures[i][j & 3],
					&pheader->fbtextures[i][j & 3],
					name,
					fbr_mask_name,
					extname,
					(byte *) (pskintype),
					pheader->skinwidth,
					pheader->skinheight,
					offset
				);

				// johnfitz
				pskintype = (daliasskintype_t *) ((byte *) (pskintype) + size);
			}

			k = j;

			for (/* */; j < 4; j++)
				pheader->gltextures[i][j & 3] =
					pheader->gltextures[i][j - k];
		}
	}

	return (void *) pskintype;
}


/*
=================
Mod_SetExtraFlags -- johnfitz -- set up extra flags that aren't in the mdl
=================
*/
void Mod_SetExtraFlags (model_t *mod)
{
	extern cvar_t r_nolerp_list;
	char *s;
	int i;

	if (!mod || !mod->name || mod->type != mod_alias)
		return;

	mod->flags &= 0xFF; // only preserve first byte

	// nolerp flag
	for (s = r_nolerp_list.string; *s; s += i + 1, i = 0)
	{
		// search forwards to the next comma or end of string
		for (i = 0; s[i] != ',' && s[i] != 0; i++) ;

		// compare it to the model name
		if (!strncmp (mod->name, s, i))
		{
			mod->flags |= MOD_NOLERP;
			break;
		}
	}

	// noshadow flag (TODO: make this a cvar list)
	if (!strcmp (mod->name, "progs/flame2.mdl") ||
			!strcmp (mod->name, "progs/flame.mdl") ||
			!strcmp (mod->name, "progs/bolt1.mdl") ||
			!strcmp (mod->name, "progs/bolt2.mdl") ||
			!strcmp (mod->name, "progs/bolt3.mdl") ||
			!strcmp (mod->name, "progs/laser.mdl"))
		mod->flags |= MOD_NOSHADOW;

	// fullbright hack (TODO: make this a cvar list)
	if (!strcmp (mod->name, "progs/flame2.mdl") ||
			!strcmp (mod->name, "progs/flame.mdl") ||
			!strcmp (mod->name, "progs/boss.mdl"))
		mod->flags |= MOD_FBRIGHTHACK;
}

/*
=================
Mod_LoadAliasModel
=================
*/
void Mod_LoadAliasModel (model_t *mod, void *buffer)
{
	int					i;
	mdl_t				*pinmodel;
	stvert_t			*pinstverts;
	dtriangle_t			*pintriangles;
	int					version;
	daliasframetype_t	*pframetype;
	daliasskintype_t	*pskintype;
	int					start, end, total;
	int					startt, endt;

	// look for an IQM to replace it
	if (Mod_FindIQMModel (mod)) return;

	startt = SDL_GetTicks ();
	start = Hunk_LowMark ();

	pinmodel = (mdl_t *) buffer;
	mod_base = (byte *) buffer; // johnfitz

	version = LittleLong (pinmodel->version);

	if (version != ALIAS_VERSION)
	{
		Host_Error ("%s has wrong version number (%i should be %i)", mod->name, version, ALIAS_VERSION);
		return;
	}

	// allocate space for a working header
	pheader = Hunk_AllocName (sizeof (aliashdr_t) + LittleLong (pinmodel->numframes) * sizeof (maliasframedesc_t), loadname);
	mod->flags = LittleLong (pinmodel->flags);
	mod->type = mod_alias;

	// endian-adjust and copy the data, starting with the alias model header
	pheader->boundingradius = LittleFloat (pinmodel->boundingradius);
	pheader->numskins = LittleLong (pinmodel->numskins);
	pheader->skinwidth = LittleLong (pinmodel->skinwidth);
	pheader->skinheight = LittleLong (pinmodel->skinheight);
	pheader->vertsperframe = LittleLong (pinmodel->numverts);
	pheader->numtris = LittleLong (pinmodel->numtris);
	pheader->numframes = LittleLong (pinmodel->numframes);
	pheader->numverts = pheader->numtris * 3;

	// validate the setup
	// Sys_Error seems a little harsh here...
	if (pheader->numframes < 1) Host_Error ("Mod_LoadAliasModel: Invalid # of frames: %d\n", pheader->numframes);
	if (pheader->numtris <= 0) Host_Error ("model %s has no triangles", mod->name);
	if (pheader->vertsperframe <= 0) Host_Error ("model %s has no vertices", mod->name);

	pheader->size = LittleFloat (pinmodel->size) * ALIAS_BASE_SIZE_RATIO;
	mod->synctype = (synctype_t) LittleLong (pinmodel->synctype);
	mod->numframes = pheader->numframes;

	for (i = 0; i < 3; i++)
	{
		pheader->scale[i] = LittleFloat (pinmodel->scale[i]);
		pheader->scale_origin[i] = LittleFloat (pinmodel->scale_origin[i]);
	}

	// load the skins
	pskintype = (daliasskintype_t *) &pinmodel[1];
	pskintype = Mod_LoadAllSkins (pheader->numskins, pskintype);

	// load base s and t vertices
	pinstverts = (stvert_t *) pskintype;

	for (i = 0; i < pheader->vertsperframe; i++)
	{
		pinstverts[i].onseam = LittleLong (pinstverts[i].onseam);
		pinstverts[i].s = LittleLong (pinstverts[i].s);
		pinstverts[i].t = LittleLong (pinstverts[i].t);
	}

	// load triangle lists
	pintriangles = (dtriangle_t *) &pinstverts[pheader->vertsperframe];

	for (i = 0; i < pheader->numtris; i++)
	{
		pintriangles[i].facesfront = LittleLong (pintriangles[i].facesfront);

		pintriangles[i].vertindex[0] = LittleLong (pintriangles[i].vertindex[0]);
		pintriangles[i].vertindex[1] = LittleLong (pintriangles[i].vertindex[1]);
		pintriangles[i].vertindex[2] = LittleLong (pintriangles[i].vertindex[2]);
	}

	// load the frames
	pheader->nummeshframes = 0;
	pheader->vertexes = 0;
	pheader->framevertexsize = 0;
	pframetype = (daliasframetype_t *) &pintriangles[pheader->numtris];

	loadmodel->mins[0] = loadmodel->mins[1] = loadmodel->mins[2] = 99999999;
	loadmodel->maxs[0] = loadmodel->maxs[1] = loadmodel->maxs[2] = -99999999;

	for (i = 0; i < pheader->numframes; i++)
	{
		aliasframetype_t frametype = LittleLong (pframetype->type);

		if (frametype == ALIAS_SINGLE)
			pframetype = (daliasframetype_t *) Mod_LoadAliasFrame (pframetype + 1, &pheader->frames[i]);
		else pframetype = (daliasframetype_t *) Mod_LoadAliasGroup (pframetype + 1, &pheader->frames[i]);
	}

	for (i = 0; i < 3; i++)
	{
		loadmodel->mins[i] = (loadmodel->mins[i] * pheader->scale[i]) + pheader->scale_origin[i];
		loadmodel->maxs[i] = (loadmodel->maxs[i] * pheader->scale[i]) + pheader->scale_origin[i];
	}

	Mod_LoadAliasBBoxes ();

	// build the draw lists
	GL_MakeAliasModelDisplayLists (pinstverts, pintriangles);

	Mod_SetExtraFlags (mod); // johnfitz

	// move the complete, relocatable alias model to the cache
	end = Hunk_LowMark ();
	total = end - start;
	Cache_Alloc (&mod->cache, total, loadname);

	if (!mod->cache.data) return;

	memcpy (mod->cache.data, pheader, total);
	Hunk_FreeToLowMark (start);

	endt = SDL_GetTicks ();

	Con_DPrintf ("%s.mdl in %i ms\n", loadname, (endt - startt));
}


// =============================================================================

/*
=================
Mod_LoadSpriteFrame
=================
*/
void *Mod_LoadSpriteFrame (void *pin, mspriteframe_t **ppframe, int framenum)
{
	dspriteframe_t		*pinframe;
	mspriteframe_t		*pspriteframe;
	int					width, height, size, origin[2];
	char				name[64];
	src_offset_t		offset; // johnfitz
	byte				*data = NULL;
	int					hunkmark, i;
	char				extname[256];

	pinframe = (dspriteframe_t *) pin;

	width = LittleLong (pinframe->width);
	height = LittleLong (pinframe->height);
	size = width * height;

	pspriteframe = (mspriteframe_t *) Hunk_Alloc (sizeof (mspriteframe_t));

	memset (pspriteframe, 0, sizeof (mspriteframe_t));

	*ppframe = pspriteframe;

	pspriteframe->width = width;
	pspriteframe->height = height;
	origin[0] = LittleLong (pinframe->origin[0]);
	origin[1] = LittleLong (pinframe->origin[1]);

	pspriteframe->up = origin[1];
	pspriteframe->down = origin[1] - height;
	pspriteframe->left = origin[0];
	pspriteframe->right = width + origin[0];

	// Image_LoadImage puts data on the hunk and it;s our responsibility to free it...
	hunkmark = Hunk_LowMark ();

	for (i = 0; ; i++)
	{
		// no more paths
		if (!texpaths[i]) break;

		sprintf (extname, "%s%s.spr_%i", texpaths[i], loadname, framenum);

		data = Image_LoadImage (extname, &width, &height);

		if (data)
		{
			// no padding; it's expected that our texture artists will create power of two images
			pspriteframe->smax = 1;
			pspriteframe->tmax = 1;

			pspriteframe->gltexture = TM_LoadImage (
				loadmodel,
				extname,
				width,
				height,
				SRC_RGBA,
				data,
				extname,
				0,
				TEXPREF_ALPHA | TEXPREF_NOPICMIP); // johnfitz -- TexMgr

			// got it
			break;
		}
	}

	// not loaded externally
	if (!data)
	{
		// johnfitz -- image might be padded
		pspriteframe->smax = (float) width / (float) TM_PadConditional (width);
		pspriteframe->tmax = (float) height / (float) TM_PadConditional (height);
		// johnfitz

		sprintf (name, "%s:frame%i", loadmodel->name, framenum);
		offset = (src_offset_t) (pinframe + 1) - (src_offset_t) mod_base; // johnfitz

		pspriteframe->gltexture = TM_LoadImage (
			loadmodel,
			name,
			width,
			height,
			SRC_INDEXED,
			(byte *) (pinframe + 1),
			loadmodel->name,
			offset,
			TEXPREF_PAD | TEXPREF_ALPHA | TEXPREF_NOPICMIP); // johnfitz -- TexMgr
	}

	// Image_LoadImage puts data on the hunk and it;s our responsibility to free it...
	Hunk_FreeToLowMark (hunkmark);

	return (void *) ((byte *) pinframe + sizeof (dspriteframe_t) + size);
}


/*
=================
Mod_LoadSpriteGroup
=================
*/
void *Mod_LoadSpriteGroup (void *pin, mspriteframe_t **ppframe, int framenum)
{
	dspritegroup_t		*pingroup;
	mspritegroup_t		*pspritegroup;
	int					i, numframes;
	dspriteinterval_t	*pin_intervals;
	float				*poutintervals;
	void				*ptemp;

	pingroup = (dspritegroup_t *) pin;

	numframes = LittleLong (pingroup->numframes);

	pspritegroup = (mspritegroup_t *) Hunk_Alloc (sizeof (mspritegroup_t) + (numframes - 1) * sizeof (pspritegroup->frames[0]));

	pspritegroup->numframes = numframes;

	*ppframe = (mspriteframe_t *) pspritegroup;

	pin_intervals = (dspriteinterval_t *) (pingroup + 1);

	poutintervals = (float *) Hunk_Alloc (numframes * sizeof (float));

	pspritegroup->intervals = poutintervals;

	for (i = 0; i < numframes; i++)
	{
		*poutintervals = LittleFloat (pin_intervals->interval);

		if (*poutintervals <= 0.0)
			Host_Error ("Mod_LoadSpriteGroup: interval<=0");

		poutintervals++;
		pin_intervals++;
	}

	ptemp = (void *) pin_intervals;

	for (i = 0; i < numframes; i++)
	{
		ptemp = Mod_LoadSpriteFrame (ptemp, &pspritegroup->frames[i], framenum * 100 + i);
	}

	return ptemp;
}


/*
=================
Mod_LoadSpriteModel
=================
*/
void Mod_LoadSpriteModel (model_t *mod, void *buffer)
{
	int					i;
	int					version;
	dsprite_t			*pin;
	msprite_t			*psprite;
	int					numframes;
	int					size;
	dspriteframetype_t	*pframetype;

	pin = (dsprite_t *) buffer;
	mod_base = (byte *) buffer; // johnfitz

	version = LittleLong (pin->version);

	if (version != SPRITE_VERSION)
		Host_Error ("%s has wrong version number "
				   "(%i should be %i)", mod->name, version, SPRITE_VERSION);

	numframes = LittleLong (pin->numframes);

	size = sizeof (msprite_t) +	(numframes - 1) * sizeof (psprite->frames);

	psprite = (msprite_t *) Hunk_Alloc (size);

	mod->cache.data = psprite;

	psprite->type = LittleLong (pin->type);
	psprite->maxwidth = LittleLong (pin->width);
	psprite->maxheight = LittleLong (pin->height);
	psprite->beamlength = LittleFloat (pin->beamlength);
	psprite->litsprite = false;
	mod->synctype = LittleLong (pin->synctype);
	psprite->numframes = numframes;

	for (i = 0; ; i++)
	{
		if (!mod->name[i]) break;

		if (mod->name[i] == '!')
		{
			psprite->litsprite = true;
			break;
		}
	}

	mod->mins[0] = mod->mins[1] = -psprite->maxwidth / 2;
	mod->maxs[0] = mod->maxs[1] = psprite->maxwidth / 2;
	mod->mins[2] = -psprite->maxheight / 2;
	mod->maxs[2] = psprite->maxheight / 2;

	// 
	// load the frames
	// 
	if (numframes < 1)
		Host_Error ("Mod_LoadSpriteModel: Invalid # of frames: %d\n", numframes);

	mod->numframes = numframes;

	pframetype = (dspriteframetype_t *) (pin + 1);

	for (i = 0; i < numframes; i++)
	{
		spriteframetype_t	frametype;

		frametype = LittleLong (pframetype->type);
		psprite->frames[i].type = frametype;

		if (frametype == SPR_SINGLE)
		{
			pframetype = (dspriteframetype_t *)
						 Mod_LoadSpriteFrame (pframetype + 1,
											  &psprite->frames[i].frameptr, i);
		}
		else
		{
			pframetype = (dspriteframetype_t *)
						 Mod_LoadSpriteGroup (pframetype + 1,
											  &psprite->frames[i].frameptr, i);
		}
	}

	mod->type = mod_sprite;
}

// =============================================================================

/*
================
Mod_Print
================
*/
void Mod_Print (void)
{
	int		i;
	model_t	*mod;

	Con_SafePrintf ("Cached models:\n"); // johnfitz -- safeprint instead of print

	for (i = 0, mod = mod_known; i < mod_numknown; i++, mod++)
	{
		Con_SafePrintf ("%8p : %s\n", mod->cache.data, mod->name); // johnfitz -- safeprint instead of print
	}

	Con_Printf ("%i models\n", mod_numknown); // johnfitz -- print the total too
}

