/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others
Copyright (C) 2007-2008 Kristian Duske

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_brush.c: brush model rendering. renamed from r_surf.c

// mh 2010-07-16 - rmq version
#include "quakedef.h"

float R_PlaneDist (mplane_t *plane, entity_t *ent);
extern cvar_t gl_fullbrights, r_drawflat; // johnfitz

gltexture_t	lightmap_textures[3][MAX_LIGHTMAPS];

byte blocklights[4][LM_BLOCK_WIDTH * LM_BLOCK_HEIGHT];

int *lm_allocated;
unsigned int *lm_buffer[3];
int lm_current = 0;

cvar_t r_instancedlight = {"r_instancedlight", "1", true};

// so that I don't need to update it in a million places if I change the conditions
qboolean R_UseInstancedLight (entity_t *ent)
{
	if (ent)
	{
		if (ent->instancedbmodel && r_instancedlight.value)// && !ent->model->lightdata)
			return true;
		else return false;
	}
	else return false;
}


/*
===============
R_TextureAnimation -- johnfitz -- added "frame" param to eliminate use of "currententity" global

mh - "currententity" no longer exists in the engine. :)

Returns the proper texture for a given time and base texture
===============
*/
texture_t *R_TextureAnimation (texture_t *base, int frame)
{
	texture_t *saved = base;
	int		relative;
	int		count;

	if (frame)
		if (base->alternate_anims)
			base = base->alternate_anims;

	if (!base->anim_total)
		return base;

	relative = (int) (cl.time * 10) % base->anim_total;
	count = 0;

	while (base->anim_min > relative || base->anim_max <= relative)
	{
		base = base->anim_next;

		if (!base) return saved;
		if (++count > 100) return saved;
	}

	return base;
}


void R_BuildLightMap (msurface_t *surf, byte *dest, int stride, int mapnum)
{
	int			smax, tmax;
	int			t;
	int			i, j, size;
	byte		*lightmap;
	int			maps;

	smax = (surf->extents[0] >> 4) + 1;
	tmax = (surf->extents[1] >> 4) + 1;
	size = smax * tmax;
	lightmap = surf->samples;

	// set to full bright if no light data
	if (!cl.worldmodel->lightdata)
	{
		for (i = 0; i < size; i++)
		{
			blocklights[0][i] = 255;
			blocklights[1][i] = 255;
			blocklights[2][i] = 255;
			blocklights[3][i] = 255;
		}
	}
	else
	{
		// clear to no light
		for (i = 0; i < size; i++)
		{
			blocklights[0][i] = 0;
			blocklights[1][i] = 0;
			blocklights[2][i] = 0;
			blocklights[3][i] = 0;
		}

		if (lightmap)
		{
			if (surf->numstyles > 1)
			{
				for (maps = 0; maps < MAX_SURFACE_STYLES && surf->styles[maps] != 255; maps++)
				{
					for (i = mapnum, j = 0; j < size; j++, i += 3)
						blocklights[maps][j] = lightmap[i];

					lightmap += size * 3;	// skip to next lightmap
				}
			}
			else
			{
				for (i = 0; i < size; i++, lightmap += 3)
				{
					blocklights[0][i] = lightmap[0];
					blocklights[1][i] = lightmap[1];
					blocklights[2][i] = lightmap[2];
					blocklights[3][i] = 255;
				}
			}
		}
	}

	stride -= smax * 4;

	for (i = 0, t = 0; i < tmax; i++, dest += stride)
	{
		for (j = 0; j < smax; j++, dest += 4, t++)
		{
			dest[0] = blocklights[0][t];
			dest[1] = blocklights[1][t];
			dest[2] = blocklights[2][t];
			dest[3] = blocklights[3][t];
		}
	}
}


/*
=============================================================

	BRUSH MODELS

=============================================================
*/

extern cvar_t r_lavaalpha;
extern cvar_t r_telealpha;
extern cvar_t r_slimealpha;
extern cvar_t r_lockalpha;

// https://trac.assembla.com/RemakeQuake/ticket/120#comment:1336
void Surf_SetAlpha (msurface_t *surf, entity_t *ent)
{
	// default opaque
	surf->wateralpha = 1.0f;

	// if the surf is sky that's all we need to do - opaque always and get out
	if (surf->flags & SURF_DRAWSKY) return;

	if (ent)
	{
		// explicit entity alpha always overrides
		if (!(ent->alpha < 1 || ent->alpha > 254))
		{
			surf->wateralpha = (float) ent->alpha / 255.0f;
			return;
		}
	}

	// evaluate alpha from surface type
	if (surf->flags & SURF_DRAWTURB)
	{
		if (!r_lockalpha.value)
		{
			if (surf->flags & SURF_DRAWLAVA)
				surf->wateralpha = r_lavaalpha.value;
			else if (surf->flags & SURF_DRAWTELE)
				surf->wateralpha = r_telealpha.value;
			else if (surf->flags & SURF_DRAWSLIME)
				surf->wateralpha = r_slimealpha.value;
			else surf->wateralpha = r_wateralpha.value;
		}
		else surf->wateralpha = r_wateralpha.value;
	}
}


/*
=================
R_DrawBrushModel
=================
*/
void R_MarkNodes (mnode_t *node)
{
	if (node->contents < 0) return;

	node->visframe = r_visframecount;

	R_MarkNodes (node->children[0]);
	R_MarkNodes (node->children[1]);
}


void R_AddSurfaceToAlphaList (msurface_t *surf, entity_t *ent, float surfalpha);
void R_AddFenceToAlphaList (msurface_t *surf, entity_t *ent, float surfalpha);


void R_ClearTextureChains (model_t *mod);
void R_ChainSurface (msurface_t *surf, entity_t *ent);
void R_DrawTextureChains (model_t *mod, entity_t *ent);

int R_ChainBrushModelSurfaces (model_t *mod, entity_t *e, qboolean enforceuniquesurfs)
{
	int i;
	msurface_t	*surf;
	float		dot;
	int r_numdrawsurfs = 0;

	// how could this possibly ever happen???  alpha was already fixed up in R_MergeBrushEntities.  but yet it does
	if (e->alpha < 1) e->alpha = 255;

	// draw it
	for (i = 0, surf = &mod->surfaces[mod->firstmodelsurface]; i < mod->nummodelsurfaces; i++, surf++)
	{
		if (surf->visframe == r_framecount && enforceuniquesurfs) continue;

		// fucking thing works again...
		dot = R_PlaneDist (surf->plane, e);

		if (((surf->flags & SURF_PLANEBACK) && (dot < -BACKFACE_EPSILON)) || (!(surf->flags & SURF_PLANEBACK) && (dot > BACKFACE_EPSILON)))
		{
			// now mark as visible in this frame
			surf->visframe = r_framecount;

			Surf_SetAlpha (surf, e);

			// note - we don't move water surfs to the alpha list yet here because we must check when drawing for
			// different lava/slime/tele/water alpha values
			if (surf->flags & SURF_DRAWFENCE)
				R_AddFenceToAlphaList (surf, e, surf->wateralpha);
			else if ((surf->wateralpha < 1) && !(surf->flags & SURF_DRAWTURB))
				R_AddSurfaceToAlphaList (surf, e, surf->wateralpha);
			else
			{
				R_ChainSurface (surf, e);
				r_numdrawsurfs++;
			}

			// correct the model (all brush surfs are now merged into the world, this is
			// only used for rebuilding and translating the verts)
			surf->model = e->model;
		}
	}

	GL_MultiplyMatrix (&e->matrix, &e->matrix, &r_world_matrix);
	return r_numdrawsurfs;
}


void R_LightBrushModel (entity_t *e, model_t *clmodel)
{
	// take an initial lighting
	e->aliasstate.shadelight[0] = 0.0f;
	e->aliasstate.shadelight[1] = 0.0f;
	e->aliasstate.shadelight[2] = 0.0f;
	e->aliasstate.shadelight[3] = 0.0f;

	// calculate dynamic lighting for bmodel if it's not an instanced model
	if (clmodel->firstmodelsurface != 0)
	{
		// store for the dlight push
		e->headnode = clmodel->nodes + clmodel->hulls[0].firstclipnode;
	}
	else if (R_UseInstancedLight (e))
	{
		// store for the dlight push
		e->headnode = clmodel->nodes;// + clmodel->hulls[0].firstclipnode;

		// now light the bmodel for real
		R_LightPoint (e, e->aliasstate.shadelight);

		// take to final scale
		VectorScale (e->aliasstate.shadelight, (r_instancedlight.value / 128.0f), e->aliasstate.shadelight);

		// because it will be added to the lightmap
		e->aliasstate.shadelight[3] = 0;
	}
	else e->headnode = NULL;
}


void R_DrawBrushModel (entity_t *e)
{
	model_t		*clmodel;
	int			r_numdrawsurfs;

	// to do - an offscreen model should probably cache anyway...
	if (R_CullBox (e->mins, e->maxs)) return;
	if (!r_drawentities.value) return;

	clmodel = e->model;
	VectorSubtract (r_refdef.vieworg, e->origin, e->modelorg);

	// hide the entity until we know that we need to draw it (based on r_numdrawsurfs)
	e->visframe = -1;
	r_numdrawsurfs = 0;

	// always set the entity matrix so that we're prepared if we ever need it
	GL_IdentityMatrix (&e->matrix);

	if (e->angles[0] || e->angles[1] || e->angles[2])
	{
		// fucking thing works again...
		vec3_t	temp;
		vec3_t	forward, right, up;

		VectorCopy (e->modelorg, temp);
		AngleVectors (e->angles, forward, right, up);

		e->modelorg[0] = DotProduct (temp, forward);
		e->modelorg[1] = -DotProduct (temp, right);
		e->modelorg[2] = DotProduct (temp, up);
	}

	GL_TranslateMatrix (&e->matrix, e->origin[0], e->origin[1], e->origin[2]);

	// don't need to negate angles[0] as it's not going through the extra negation in R_RotateForEntity
	if (e->angles[1]) GL_RotateMatrix (&e->matrix, e->angles[1], 0, 0, 1);
	if (e->angles[0]) GL_RotateMatrix (&e->matrix, e->angles[0], 0, 1, 0);
	if (e->angles[2]) GL_RotateMatrix (&e->matrix, e->angles[2], 1, 0, 0);

	if (e->scale != 16)
		GL_ScaleMatrix (&e->matrix, (float) e->scale / 16.0f, (float) e->scale / 16.0f, (float) e->scale / 16.0f);

	// calculate entity local space for dlight transforms
	GL_IdentityMatrix (&e->local);

	// don't need to negate angles[0] as it's not going through the extra negation in R_RotateForEntity
	if (e->angles[2]) GL_RotateMatrix (&e->local, -e->angles[2], 1, 0, 0);
	if (e->angles[0]) GL_RotateMatrix (&e->local, -e->angles[0], 0, 1, 0);
	if (e->angles[1]) GL_RotateMatrix (&e->local, -e->angles[1], 0, 0, 1);

	GL_TranslateMatrix (&e->local, -e->origin[0], -e->origin[1], -e->origin[2]);

	// mark dynamic lights on the model
	// R_PushDlights (clmodel->nodes + clmodel->hulls[0].firstclipnode, e);

	R_LightBrushModel (e, clmodel);

	R_ClearTextureChains (e->model);
	r_numdrawsurfs = R_ChainBrushModelSurfaces (e->model, e, false);

	// now mark that we got regular surfaces for this entity
	// (to do - if the entity was merged to the world we don't need to draw it separately)
	if (r_numdrawsurfs)
	{
		// Con_Printf ("Drawing %i surfs on %s\n", r_numdrawsurfs, clmodel->name);
		e->visframe = r_framecount;
		R_DrawTextureChains (e->model, e);
	}
}


/*
=============================================================

	LIGHTMAPS

=============================================================
*/


void LM_InitBlock (void)
{
	memset (lm_allocated, 0, LM_BLOCK_WIDTH * sizeof (int));
}


void LM_UploadBlock (void)
{
	int j;

	for (j = 0; j < 3; j++)
	{
		if (!lightmap_textures[j][lm_current].texnum)
		{
			glGenTextures (1, &lightmap_textures[j][lm_current].texnum);

			GL_BindTexture (GL_DONT_CARE, &lightmap_textures[j][lm_current]);

			glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

			glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA8, LM_BLOCK_WIDTH, LM_BLOCK_HEIGHT, 0, GL_RGBA, GL_UNSIGNED_BYTE, lm_buffer[j]);
		}
		else
		{
			GL_BindTexture (GL_DONT_CARE, &lightmap_textures[j][lm_current]);
			glTexSubImage2D (GL_TEXTURE_2D, 0, 0, 0, LM_BLOCK_WIDTH, LM_BLOCK_HEIGHT, GL_RGBA, GL_UNSIGNED_BYTE, lm_buffer[j]);
		}

		// clear the buffer
		memset (lm_buffer[j], 0, LM_BLOCK_WIDTH * LM_BLOCK_HEIGHT * 4);
	}

	if ((++lm_current) >= MAX_LIGHTMAPS) Sys_Error ("LM_Uploadblock : MAX_LIGHTMAPS exceeded");
}


qboolean LM_AllocBlock (int w, int h, int *x, int *y)
{
	int		i, j;
	int		best, best2;

	best = LM_BLOCK_HEIGHT;

	for (i = 0; i < LM_BLOCK_WIDTH - w; i++)
	{
		best2 = 0;

		for (j = 0; j < w; j++)
		{
			if (lm_allocated[i + j] >= best) break;
			if (lm_allocated[i + j] > best2) best2 = lm_allocated[i + j];
		}

		if (j == w)
		{
			// this is a valid spot
			*x = i;
			*y = best = best2;
		}
	}

	if (best + h > LM_BLOCK_HEIGHT)
		return false;

	for (i = 0; i < w; i++)
		lm_allocated[*x + i] = best + h;

	return true;
}


/*
========================
GL_CreateSurfaceLightmap
========================
*/
void GL_CreateSurfaceLightmap (msurface_t *surf)
{
	int j;
	unsigned int *base;

	if (surf->flags & SURF_DRAWSKY) return;
	if (surf->flags & SURF_DRAWTURB) return;

	surf->smax = (surf->extents[0] >> 4) + 1;
	surf->tmax = (surf->extents[1] >> 4) + 1;

	if (!LM_AllocBlock (surf->smax, surf->tmax, &surf->light_s, &surf->light_t))
	{
		LM_UploadBlock ();
		LM_InitBlock ();

		if (!LM_AllocBlock (surf->smax, surf->tmax, &surf->light_s, &surf->light_t))
			Sys_Error ("GL_CreateSurfaceLightmap : consecutive calls to LM_AllocBlock failed");
	}

	surf->lightmaptexturenum = lm_current;
	surf->numstyles = 0;

	for (j = 0; j < MAX_SURFACE_STYLES && surf->styles[j] != 255; j++, surf->numstyles++);

	if (surf->numstyles > 1)
	{
		for (j = 0; j < 3; j++)
		{
			base = lm_buffer[j] + (surf->light_t * LM_BLOCK_WIDTH + surf->light_s);
			R_BuildLightMap (surf, (byte *) base, LM_BLOCK_WIDTH * 4, j);
		}
	}
	else
	{
		base = lm_buffer[0] + (surf->light_t * LM_BLOCK_WIDTH + surf->light_s);
		R_BuildLightMap (surf, (byte *) base, LM_BLOCK_WIDTH * 4, 0);
	}
}


/*
================
GL_BuildPolygonForSurface -- called at level load time

renamed this because it wasn't actually building a display list!
================
*/
void GL_BuildPolygonForSurface (msurface_t *surf, model_t *mod)
{
}

/*
==================
GL_BuildLightmaps -- called at level load time

Builds the lightmap texture
with all the surfaces from all brush models
==================
*/
void GL_DeleteAllLightmaps (void)
{
	int i, j;

	for (i = 0; i < MAX_LIGHTMAPS; i++)
	{
		for (j = 0; j < 3; j++)
		{
			glDeleteTextures (1, &lightmap_textures[j][i].texnum);
			memset (&lightmap_textures[j][i], 0, sizeof (gltexture_t));
		}
	}
}


int Surf_TextureSort (msurface_t *s1, msurface_t *s2)
{
	return s1->texinfo->texture - s2->texinfo->texture;
}


int Surf_NumberSort (msurface_t *s1, msurface_t *s2)
{
	return s1->surfnum - s2->surfnum;
}

void Mod_SetCorrectBBox (model_t *mod);

void GL_BuildLightmaps (void)
{
	int i, j;
	model_t	*m;
	int hunkmark = Hunk_LowMark ();

	// this can happen on first load so prevent it
	if (!cl.worldmodel) return;

	// set up our temp pointers
	// these are OK to put on the hunk even if we restart video while a map is running because they can be released
	lm_buffer[0] = (unsigned int *) Hunk_Alloc (LM_BLOCK_WIDTH * LM_BLOCK_HEIGHT * sizeof (unsigned int));
	lm_buffer[1] = (unsigned int *) Hunk_Alloc (LM_BLOCK_WIDTH * LM_BLOCK_HEIGHT * sizeof (unsigned int));
	lm_buffer[2] = (unsigned int *) Hunk_Alloc (LM_BLOCK_WIDTH * LM_BLOCK_HEIGHT * sizeof (unsigned int));
	lm_allocated = (int *) Hunk_Alloc (LM_BLOCK_WIDTH * sizeof (int));

	memset (lm_allocated, 0, LM_BLOCK_WIDTH * sizeof (int));
	lm_current = 0;
	r_framecount = 1; // no dlightcache

	for (j = 1; j < MAX_MODELS; j++)
	{
		msurface_t *surf = NULL;

		if (!(m = cl.model_precache[j])) break;
		if (m->type != mod_brush) continue;

		Mod_SetCorrectBBox (m);

		if (m->name[0] == '*') continue;

		// order surfs by texture
		qsort (m->surfaces, m->numsurfaces, sizeof (msurface_t), (sortfunc_t) Surf_TextureSort);

		for (i = 0, surf = m->surfaces; i < m->numsurfaces; i++, surf++)
			GL_CreateSurfaceLightmap (surf);

		// restore original order
		qsort (m->surfaces, m->numsurfaces, sizeof (msurface_t), (sortfunc_t) Surf_NumberSort);
	}

	LM_UploadBlock ();
	Hunk_FreeToLowMark (hunkmark);

	// clear down any lightmap textures which were unused
	for (i = lm_current; i < MAX_LIGHTMAPS; i++)
	{
		for (j = 0; j < 3; j++)
		{
			glDeleteTextures (1, &lightmap_textures[j][i].texnum);
			memset (&lightmap_textures[j][i], 0, sizeof (gltexture_t));
		}
	}
}


void R_DumpLightmap (int lmnum, byte **getbuffer, byte *outbuffer)
{
	int style;
	char lmname[256];

	GL_BindTexture (GL_DONT_CARE, &lightmap_textures[0][lmnum]);
	glGetTexImage (GL_TEXTURE_2D, 0, GL_RGBA, GL_UNSIGNED_BYTE, getbuffer[0]);

	GL_BindTexture (GL_DONT_CARE, &lightmap_textures[1][lmnum]);
	glGetTexImage (GL_TEXTURE_2D, 0, GL_RGBA, GL_UNSIGNED_BYTE, getbuffer[1]);

	GL_BindTexture (GL_DONT_CARE, &lightmap_textures[2][lmnum]);
	glGetTexImage (GL_TEXTURE_2D, 0, GL_RGBA, GL_UNSIGNED_BYTE, getbuffer[2]);

	// switch the lightmaps around to something that makes sense to us mere mortals
	for (style = 0; style < 4; style++)
	{
		int i, size;
		byte *red, *green, *blue, *dest;

		// because we set styles up in reverse order we need to reverse them again to get the correct order for dumping
		sprintf (lmname, "lightmap_%i_style_%i.tga", lmnum, 3 - style);
		memset (outbuffer, 0, LM_BLOCK_WIDTH * LM_BLOCK_HEIGHT * 3);

		red = getbuffer[0] + style;
		green = getbuffer[1] + style;
		blue = getbuffer[2] + style;
		dest = outbuffer;
		size = LM_BLOCK_WIDTH * LM_BLOCK_HEIGHT * 4;

		for (i = 0; i < size; i += 4, dest += 3)
		{
			// TGAs are BGR
			dest[2] = red[i];
			dest[1] = green[i];
			dest[0] = blue[i];
		}

		Image_WriteTGA (lmname, outbuffer, LM_BLOCK_WIDTH, LM_BLOCK_HEIGHT, 24, true);
	}
}


void R_LightmapDump_f (void)
{
	int i;
	int hunkmark;
	byte *lmGetTexBuffer[3];
	byte *lmOutputBuffer;

	hunkmark = Hunk_LowMark ();

	lmGetTexBuffer[0] = Hunk_Alloc (LM_BLOCK_WIDTH * LM_BLOCK_HEIGHT * 4);
	lmGetTexBuffer[1] = Hunk_Alloc (LM_BLOCK_WIDTH * LM_BLOCK_HEIGHT * 4);
	lmGetTexBuffer[2] = Hunk_Alloc (LM_BLOCK_WIDTH * LM_BLOCK_HEIGHT * 4);
	lmOutputBuffer = Hunk_Alloc (LM_BLOCK_WIDTH * LM_BLOCK_HEIGHT * 3);

	for (i = 0; i < MAX_LIGHTMAPS; i++)
	{
		if (lightmap_textures[0][i].texnum &&
			lightmap_textures[1][i].texnum &&
			lightmap_textures[2][i].texnum)
		{
			// all 3 textures must be present to do the dump
			R_DumpLightmap (i, lmGetTexBuffer, lmOutputBuffer);
		}
	}

	Hunk_FreeToLowMark (hunkmark);
}

